# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

from dataclasses import asdict
from rich import print
from typing import Any, Dict, List, Optional, Tuple
import sys

from PHX.from_WUFI_XML import wufi_file_schema as wufi_xml

from PHX.model.building import PhxBuilding, PhxZone
from PHX.model.certification import PhxPhiusCertification, PhxPhBuildingData
from PHX.model.constructions import (
    PhxConstructionOpaque,
    PhxConstructionWindow,
    PhxLayer,
    PhxMaterial,
    PhxWindowFrameElement,
)
from PHX.model.components import (
    PhxComponentAperture,
    PhxComponentOpaque,
    PhxComponentThermalBridge,
    PhxApertureElement,
)
from PHX.model.enums.building import (
    ComponentFaceType,
    ComponentExposureExterior,
    ComponentFaceOpacity,
    ComponentColor,
    ThermalBridgeType,
    SpecificHeatCapacity,
)
from PHX.model.enums import hvac as hvac_enums
from PHX.model.enums.phius_certification import (
    PhiusCertificationProgram,
    PhiusCertificationBuildingCategoryType,
    PhiusCertificationBuildingUseType,
    PhiusCertificationBuildingStatus,
    PhiusCertificationBuildingType,
)
from PHX.model.enums.phx_site import (
    SiteSelection,
    SiteClimateSelection,
    SiteEnergyFactorSelection,
)
from PHX.model.geometry import PhxVertix, PhxPolygon, PhxVector, PhxPlane
from PHX.model.enums.foundations import (
    FoundationType,
    CalculationSetting,
    PerimeterInsulationPosition,
)
from PHX.model.ground import (
    PhxFoundation,
    PhxHeatedBasement,
    PhxUnHeatedBasement,
    PhxSlabOnGrade,
    PhxVentedCrawlspace,
)
from PHX.model.spaces import PhxSpace
from PHX.model.shades import PhxWindowShade
from PHX.model.schedules.ventilation import (
    PhxScheduleVentilation,
    Vent_UtilPeriods,
    Vent_OperatingPeriod,
)
from PHX.model.schedules.occupancy import PhxScheduleOccupancy
from PHX.model.schedules.lighting import PhxScheduleLighting
from PHX.model.phx_site import (
    PhxSite,
    PhxClimate,
    PhxGround,
    PhxSiteEnergyFactors,
    PhxPEFactor,
    PhxCO2Factor,
)
from PHX.model.project import (
    PhxProject,
    PhxProjectData,
    ProjectData_Agent,
    PhxVariant,
    WufiPlugin,
)
from PHX.model.hvac.collection import PhxZoneCoverage
from PHX.model.programs.occupancy import PhxProgramOccupancy
from PHX.model.programs.ventilation import PhxProgramVentilation
from PHX.model.hvac.ventilation import (
    PhxDeviceVentilator,
    AnyPhxExhaustVent,
    PhxExhaustVentilatorRangeHood,
    PhxExhaustVentilatorDryer,
    PhxExhaustVentilatorUserDefined,
)
from PHX.model.hvac.heating import (
    PhxHeaterBoilerFossil,
    PhxHeaterBoilerWood,
    PhxHeaterDistrictHeat,
    PhxHeaterElectric,
    PhxHeaterHeatPumpAnnual,
    PhxHeaterHeatPumpCombined,
    PhxHeaterHeatPumpMonthly,
    PhxHeaterHeatPumpHotWater,
    PhxHeaterHeatPump,
    PhxHeaterBoiler,
)
from PHX.model.hvac.water import PhxHotWaterTank
from PHX.model.hvac.renewable_devices import PhxDevicePhotovoltaic
from PHX.model.enums.elec_equip import ElectricEquipmentType
from PHX.model.elec_equip import (
    PhxElectricalDevice,
    PhxDeviceDishwasher,
    PhxDeviceCooktop,
    PhxDeviceClothesWasher,
    PhxDeviceClothesDryer,
    PhxDeviceRefrigerator,
    PhxDeviceFreezer,
    PhxDeviceFridgeFreezer,
    PhxDeviceCooktop,
    PhxDeviceCustomElec,
    PhxDeviceMEL,
    PhxDeviceLightingInterior,
    PhxDeviceLightingExterior,
    PhxDeviceLightingGarage,
    PhxDeviceCustomLighting,
    PhxDeviceCustomMEL,
)


# ----------------------------------------------------------------------
# -- Conversion Function


def as_phx_obj(_model, _schema_name, **kwargs) -> Any:
    """Find the right class-builder from the module and pass along the data to it.

    Args:
        _model (Any): The Pydantic (BaseModel) item to be converted.
        _schema_name ([type]): The name of the class-builder to be used.
        kwargs: Any additional arguments to be passed to the class-builder.
    Returns:
        A new PHX object built from the input data
    """
    builder = getattr(sys.modules[__name__], f"_{_schema_name}")
    return builder(_model, **kwargs)


# ----------------------------------------------------------------------
# -- Project


def _PhxProject(_model: wufi_xml.WUFIplusProject) -> PhxProject:
    phx_obj = PhxProject()
    phx_obj.data_version = _model.DataVersion
    phx_obj.unit_system = _model.UnitSystem
    phx_obj.program_version = _model.ProgramVersion
    phx_obj.scope = _model.Scope
    phx_obj.visualized_geometry = _model.DimensionsVisualizedGeometry

    phx_obj.project_data = as_phx_obj(_model.ProjectData, "PhxProjectData")

    # ----------------------------------------------------------------------
    # -- Build all the type collections first
    for window_type_data in _model.WindowTypes:
        new_window = as_phx_obj(window_type_data, "PhxConstructionWindow")
        # -- Be sure to use the identifier as the key so the Component
        # -- lookup works properly. We don't use the name here since
        # -- wufi doesn't enforce unique names like HB does.
        phx_obj.add_new_window_type(new_window, _key=new_window.identifier)

    for assembly_data in _model.Assemblies:
        new_opaque: PhxConstructionOpaque = as_phx_obj(
            assembly_data, "PhxConstructionOpaque"
        )
        # -- Be sure to use the identifier as the key so the Component
        # -- lookup works properly. We don't use the name here since
        # -- wufi doesn't enforce unique names like HB does.
        phx_obj.add_assembly_type(new_opaque, _key=new_opaque.identifier)

    for shade_type_data in _model.SolarProtectionTypes or []:
        new_shade: PhxWindowShade = as_phx_obj(shade_type_data, "PhxWindowShade")
        # -- Be sure to use the identifier as the key so the Component
        # -- lookup works properly. We don't use the name here since
        # -- wufi doesn't enforce unique names like HB does.
        phx_obj.add_new_shade_type(new_shade, _key=new_shade.identifier)

    for vent_pattern_data in _model.UtilisationPatternsVentilation:
        new_pattern = as_phx_obj(vent_pattern_data, "PhxScheduleVentilation")
        phx_obj.utilization_patterns_ventilation.add_new_util_pattern(new_pattern)

    for vent_pattern_data in _model.UtilizationPatternsPH:
        new_pattern = as_phx_obj(vent_pattern_data, "PhxScheduleOccupancy")
        phx_obj.utilization_patterns_occupancy.add_new_util_pattern(new_pattern)

    # ----------------------------------------------------------------------
    # -- Build all the actual Variants
    for variant_dict in _model.Variants:
        new_variant = as_phx_obj(variant_dict, "PhxVariant", _phx_project_host=phx_obj)
        phx_obj.add_new_variant(new_variant)

    return phx_obj


def _PhxProjectData(_model: wufi_xml.ProjectData) -> PhxProjectData:
    phx_obj = PhxProjectData()
    # phx_obj.project_date = _model.Date_Project
    phx_obj.owner_is_client = _model.OwnerIsClient
    phx_obj.year_constructed = int(_model.Year_Construction)
    phx_obj.image = _model.WhiteBackgroundPictureBuilding

    phx_obj.customer = ProjectData_Agent(
        _model.Customer_Name,
        _model.Customer_Street,
        _model.Customer_Locality,
        _model.Customer_PostalCode,
        _model.Customer_Tel,
        _model.Customer_Email,
    )
    phx_obj.building = ProjectData_Agent(
        _model.Building_Name,
        _model.Building_Street,
        _model.Building_Locality,
        _model.Building_PostalCode,
    )
    phx_obj.owner = ProjectData_Agent(
        _model.Owner_Name,
        _model.Owner_Street,
        _model.Owner_Locality,
        _model.Owner_PostalCode,
    )
    phx_obj.designer = ProjectData_Agent(
        _model.Responsible_Name,
        _model.Responsible_Street,
        _model.Responsible_Locality,
        _model.Responsible_PostalCode,
        _model.Responsible_Tel,
        _model.Responsible_LicenseNr,
        _model.Responsible_Email,
    )

    return phx_obj


# ----------------------------------------------------------------------
# -- Envelope Types


def _PhxConstructionWindow(_t: wufi_xml.WindowType) -> PhxConstructionWindow:
    phx_obj = PhxConstructionWindow()
    phx_obj.id_num = _t.IdentNr
    phx_obj.identifier = str(_t.IdentNr)
    phx_obj.display_name = _t.Name
    phx_obj.use_detailed_uw = _t.Uw_Detailed
    phx_obj.use_detailed_frame = _t.GlazingFrameDetailed

    phx_obj.u_value_window = _t.U_Value
    phx_obj.u_value_glass = _t.U_Value_Glazing
    phx_obj.u_value_frame = _t.U_Value_Frame

    phx_obj.glass_mean_emissivity = _t.MeanEmissivity
    phx_obj.glass_g_value = _t.g_Value

    phx_obj.frame_top = PhxWindowFrameElement(
        float(_t.Frame_Width_Top),
        float(_t.Frame_U_Top),
        float(_t.Glazing_Psi_Top),
        float(_t.Frame_Psi_Top),
    )
    phx_obj.frame_right = PhxWindowFrameElement(
        float(_t.Frame_Width_Right),
        float(_t.Frame_U_Right),
        float(_t.Glazing_Psi_Right),
        float(_t.Frame_Psi_Right),
    )
    phx_obj.frame_bottom = PhxWindowFrameElement(
        float(_t.Frame_Width_Bottom),
        float(_t.Frame_U_Bottom),
        float(_t.Glazing_Psi_Bottom),
        float(_t.Frame_Psi_Bottom),
    )
    phx_obj.frame_left = PhxWindowFrameElement(
        float(_t.Frame_Width_Left),
        float(_t.Frame_U_Left),
        float(_t.Glazing_Psi_Left),
        float(_t.Frame_Psi_Left),
    )

    return phx_obj


def _PhxConstructionOpaque(_data: wufi_xml.Assembly) -> PhxConstructionOpaque:
    phx_obj = PhxConstructionOpaque()
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.display_name = _data.Name
    phx_obj.layer_order = _data.Order_Layers
    phx_obj.grid_kind = _data.Grid_Kind
    for layer in _data.Layers:
        new_layer = as_phx_obj(layer, "PhxLayer")
        phx_obj.layers.append(new_layer)

    return phx_obj


def _PhxLayer(_data: wufi_xml.Layer) -> PhxLayer:
    phx_obj = PhxLayer()
    phx_obj.thickness_m = _data.Thickness
    new_mat = as_phx_obj(_data.Material, "PhxMaterial")
    phx_obj.add_material(new_mat)

    return phx_obj


def _PhxMaterial(_data: wufi_xml.Material) -> PhxMaterial:
    phx_obj = PhxMaterial()
    phx_obj.display_name = _data.Name
    phx_obj.conductivity = _data.ThermalConductivity
    phx_obj.density = _data.BulkDensity
    phx_obj.porosity = _data.Porosity
    phx_obj.heat_capacity = _data.HeatCapacity
    phx_obj.water_vapor_resistance = _data.WaterVaporResistance
    phx_obj.reference_water = _data.ReferenceWaterContent
    return phx_obj


def _PhxWindowShade(_data: wufi_xml.SolarProtectionType) -> PhxWindowShade:
    phx_obj = PhxWindowShade()
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.display_name = _data.Name
    phx_obj.operation_mode = _data.OperationMode
    phx_obj.reduction_factor = _data.MaxRedFactorRadiation or 1.0
    return phx_obj


# ----------------------------------------------------------------------
# -- Utilization Patterns


def _PhxScheduleVentilation(
    _data: wufi_xml.UtilizationPatternVent,
) -> PhxScheduleVentilation:
    phx_obj = PhxScheduleVentilation()

    phx_obj.id_num = _data.IdentNr
    phx_obj.name = _data.Name
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.operating_days = _data.OperatingDays
    phx_obj.operating_weeks = _data.OperatingWeeks
    phx_obj.operating_periods.high.period_operating_hours = _data.Maximum_DOS
    phx_obj.operating_periods.high.period_operation_speed = _data.Maximum_PDF
    phx_obj.operating_periods.standard.period_operating_hours = _data.Standard_DOS
    phx_obj.operating_periods.standard.period_operation_speed = _data.Standard_PDF
    phx_obj.operating_periods.basic.period_operating_hours = _data.Basic_DOS
    phx_obj.operating_periods.basic.period_operation_speed = _data.Basic_PDF
    phx_obj.operating_periods.minimum.period_operating_hours = _data.Minimum_DOS
    phx_obj.operating_periods.minimum.period_operation_speed = _data.Minimum_PDF
    return phx_obj


def _PhxScheduleOccupancy(_data: wufi_xml.UtilizationPattern) -> PhxScheduleOccupancy:
    phx_obj = PhxScheduleOccupancy()
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.display_name = _data.Name
    phx_obj.start_hour = _data.BeginUtilization
    phx_obj.end_hour = _data.EndUtilization
    phx_obj.annual_utilization_days = _data.AnnualUtilizationDays
    phx_obj.relative_utilization_factor = _data.RelativeAbsenteeism

    return phx_obj


# ----------------------------------------------------------------------
# -- Variants


def _WufiPlugin(_model: wufi_xml.Plugin) -> WufiPlugin:
    phx_obj = WufiPlugin()
    if not _model:
        return phx_obj

    phx_obj.insert_plugin = bool(_model.InsertPlugIn)
    phx_obj.name_dll = _model.Name_dll
    phx_obj.status_plugin = _model.StatusPlugIn
    return phx_obj


def _PhxVariant(_data: wufi_xml.Variant, _phx_project_host: PhxProject) -> PhxVariant:
    phx_obj = PhxVariant()

    phx_obj.id_num = _data.IdentNr
    phx_obj.name = _data.Name
    phx_obj.remarks = _data.Remarks
    phx_obj.plugin = as_phx_obj(_data.PlugIn, "WufiPlugin")
    phx_obj.phius_cert = as_phx_obj(_data.PassivehouseData, "PhxPhiusCertification")
    phx_obj.site = as_phx_obj(_data.ClimateLocation, "PhxSite")

    phx_obj.building = as_phx_obj(
        (_data.Building, _data.Graphics_3D),
        "PhxBuilding",
        _phx_project_host=_phx_project_host,
    )

    for system_data_dict in _data.HVAC.Systems:
        phx_obj.mech_systems.display_name = system_data_dict.Name
        phx_obj.mech_systems.id_num = system_data_dict.IdentNr
        phx_obj.mech_systems.zone_coverage = as_phx_obj(
            system_data_dict.ZonesCoverage[0], "PhxZoneCoverage"
        )

        for device_data_dict in system_data_dict.Devices:
            new_device = as_phx_obj(device_data_dict, "PhxMechanicalDevice")
            phx_obj.mech_systems.add_new_mech_device(new_device.identifier, new_device)

    return phx_obj


# ----------------------------------------------------------------------
# -- Building & Geometry


def _get_compo_data_by_type(
    _data: List[wufi_xml.Component],
) -> Tuple[List[wufi_xml.Component], List[wufi_xml.Component]]:
    """Helper function to sort the component dictionaries by type.

    Args:
        _data (List[Dict]): The list of component dictionaries.
    Returns:
        Tuple[List[Dict], List[Dict]]: The sorted lists of component dictionaries.
        - [0] = List of opaque component dicts
        - [1] = List of aperture components dicts
    """

    ap_dicts: List[wufi_xml.Component] = []
    opaque_dicts: List[wufi_xml.Component] = []
    for d in _data:
        if ComponentFaceOpacity(d.Type) == ComponentFaceOpacity.TRANSPARENT:
            ap_dicts.append(d)
        else:
            opaque_dicts.append(d)
    return opaque_dicts, ap_dicts


def _PhxBuilding(
    _data: Tuple[wufi_xml.Building, wufi_xml.Graphics_3D],
    _phx_project_host: PhxProject,
) -> PhxBuilding:
    phx_obj = PhxBuilding()

    # ------------------------------------------------------------------
    # -- First, build all the vertices and polygons needed by the Components
    bldg_data, geom_data = _data
    vertix_dict: Dict[int, PhxVertix] = {
        v.IdentNr: as_phx_obj(v, "PhxVertix") for v in geom_data.Vertices
    }
    polygon_dict: Dict[int, PhxPolygon] = {
        v.IdentNr: as_phx_obj(v, "PhxPolygon", _vertix_dict=vertix_dict)
        for v in geom_data.Polygons
    }

    # ------------------------------------------------------------------
    # -- Build the Components
    opaque_compo_dicts, ap_compo_dicts = _get_compo_data_by_type(bldg_data.Components)

    # -- Build all of the Aperture Components first
    aperture_dict: Dict[Tuple[int, ...], PhxComponentAperture] = {}
    for component_dict in ap_compo_dicts:
        ap_component: PhxComponentAperture = as_phx_obj(
            component_dict,
            "PhxComponentAperture",
            _polygons=polygon_dict,
            _window_types=_phx_project_host.window_types,
        )
        aperture_dict[ap_component.polygon_ids_sorted] = ap_component

    # -- Build all of the Opaque Components
    for component_dict in opaque_compo_dicts:
        op_component: PhxComponentOpaque = as_phx_obj(
            component_dict,
            "PhxComponentOpaque",
            _polygons=polygon_dict,
            _assembly_types=_phx_project_host.assembly_types,
            _aperture_dict=aperture_dict,
        )
        phx_obj.add_component(op_component)

    # ------------------------------------------------------------------
    # -- Build and add the Zones
    for zone_data_dict in bldg_data.Zones:
        new_zone = as_phx_obj(
            zone_data_dict, "PhxZone", _phx_project_host=_phx_project_host
        )
        phx_obj.add_zone(new_zone)

    return phx_obj


def _PhxVertix(_data: wufi_xml.Vertix) -> PhxVertix:
    phx_obj = PhxVertix()
    phx_obj.id_num = _data.IdentNr
    phx_obj.x = _data.X
    phx_obj.y = _data.Y
    phx_obj.z = _data.Z
    return phx_obj


def _PhxPolygon(
    _data: wufi_xml.Polygon, _vertix_dict: Dict[int, PhxVertix] = {}
) -> PhxPolygon:
    surface_normal = PhxVector(
        float(_data.NormalVectorX),
        float(_data.NormalVectorY),
        float(_data.NormalVectorZ),
    )

    # ----------------------------------------------------------------------
    plane = PhxPlane(
        normal_vector=surface_normal,
        origin=PhxVertix(0, 0, 0),  # temp. value
        x=PhxVector(0, 0, 0),  # temp. value
        y=PhxVector(0, 0, 0),  # temp. value
    )
    # ----------------------------------------------------------------------

    phx_obj = PhxPolygon(
        _display_name=str(_data.IdentNr),
        _area=None,
        _center=None,
        normal_vector=plane.normal_vector,
        plane=plane,
    )
    phx_obj.id_num = _data.IdentNr
    phx_obj._vertices = [_vertix_dict[_.IdentNr] for _ in _data.IdentNrPoints]
    if _data.IdentNrPolygonsInside:
        for _ in _data.IdentNrPolygonsInside:
            phx_obj.child_polygon_ids.append(_.IdentNr)

    # -- Figure out the Polygon Plane's x, y, and origin values based on the vertices
    phx_obj.plane.origin = phx_obj.vertices[0]
    phx_obj.plane.x = PhxVector.from_2_points(phx_obj.vertices[0], phx_obj.vertices[1])
    phx_obj.plane.y = phx_obj.plane.x.rotate_around(phx_obj.plane.normal_vector, 90)

    # -- Make sure to do this, otherwise area won't work properly.
    phx_obj.plane.x.unitize()
    phx_obj.plane.y.unitize()

    return phx_obj


def _PhxComponentAperture(
    _data: wufi_xml.Component,
    _polygons: Dict[int, PhxPolygon],
    _window_types: Dict[str, PhxConstructionWindow],
) -> PhxComponentAperture:
    phx_obj = PhxComponentAperture(_host=None)  # type: ignore

    phx_obj.display_name = _data.Name
    phx_obj.face_type = ComponentFaceType.WINDOW
    phx_obj.face_opacity = ComponentFaceOpacity(_data.Type)
    phx_obj.color_interior = ComponentColor(_data.IdentNrColorI)
    phx_obj.color_exterior = ComponentColor(_data.IdentNrColorE)
    phx_obj.exposure_interior = _data.InnerAttachment
    phx_obj.exposure_exterior = ComponentExposureExterior(_data.OuterAttachment)

    if _data.IdentNr_ComponentInnerSurface:
        phx_obj.interior_attachment_id = _data.IdentNr_ComponentInnerSurface

    if _data.IdentNrWindowType:
        phx_obj.window_type = _window_types[str(_data.IdentNrWindowType)]
        phx_obj.window_type.id_num_shade = _data.IdentNrSolarProtection

    phx_obj.install_depth = _data.DepthWindowReveal
    phx_obj.default_monthly_shading_correction_factor = (
        _data.DefaultCorrectionShadingMonth
    )

    for i, poly_id in enumerate(_data.IdentNrPolygons):
        polygon = _polygons[poly_id.IdentNr]
        new_ap_element = PhxApertureElement(_host=phx_obj)
        new_ap_element.display_name = f"{phx_obj.display_name} - {i}"
        new_ap_element.polygon = polygon  # TODO: make a to_rect method for polygons....
        phx_obj.add_element(new_ap_element)

    return phx_obj


def _PhxComponentOpaque(
    _data: wufi_xml.Component,
    _polygons: Dict[int, PhxPolygon],
    _assembly_types: Dict[str, PhxConstructionOpaque],
    _aperture_dict: Dict[Tuple[int, ...], PhxComponentAperture],
) -> PhxComponentOpaque:
    # -- Build either an opaque or transparent component
    phx_obj = PhxComponentOpaque()

    phx_obj._id_num = _data.IdentNr
    phx_obj.display_name = _data.Name
    phx_obj.face_opacity = ComponentFaceOpacity(_data.Type)
    phx_obj.color_interior = ComponentColor(_data.IdentNrColorI)
    phx_obj.color_exterior = ComponentColor(_data.IdentNrColorE)
    phx_obj.exposure_interior = _data.InnerAttachment
    phx_obj.exposure_exterior = ComponentExposureExterior(_data.OuterAttachment)

    if _data.IdentNr_ComponentInnerSurface:
        phx_obj.interior_attachment_id = _data.IdentNr_ComponentInnerSurface

    phx_obj.add_polygons([_polygons[i.IdentNr] for i in _data.IdentNrPolygons])
    if phx_obj.polygons:
        face_normal_angle_off_vertical = phx_obj.polygons[0].angle_from_horizontal
        phx_obj.face_type = ComponentFaceType.by_angle(face_normal_angle_off_vertical)

    # -- Add the Apertures
    # -- For each polygon, see if it has any children polygons? If so, find the
    # -- aperture element associated with the child polygon, and add it to the component.
    for poly in phx_obj.polygons:
        for child_id in poly.child_polygon_ids:
            for k, v in _aperture_dict.items():
                if child_id in k:
                    phx_obj.add_aperture(v)

    if _data.IdentNrAssembly:
        phx_obj.assembly_type_id_num = _data.IdentNrAssembly
        if phx_obj.assembly_type_id_num != -1:
            # -- -1 indicates a shade component without an assembly type
            phx_obj.assembly = _assembly_types[str(_data.IdentNrAssembly)]

    return phx_obj


def _PhxComponentThermalBridge(
    _data: wufi_xml.ThermalBridge,
) -> PhxComponentThermalBridge:
    phx_obj = PhxComponentThermalBridge()

    phx_obj.quantity = 1.0
    phx_obj.display_name = _data.Name
    phx_obj.group_number = ThermalBridgeType(_data.Type * -1)
    phx_obj.length = _data.Length
    phx_obj.psi_value = _data.PsiValue

    return phx_obj


# ----------------------------------------------------------------------
# -- Foundations


def _PhxHeatedBasement(_data: wufi_xml.FoundationInterface) -> PhxHeatedBasement:
    phx_obj = PhxHeatedBasement()
    phx_obj.display_name = _data.Name
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.floor_slab_area_m2 = _data.FloorSlabArea
    phx_obj.floor_slab_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.floor_slab_u_value = _data.U_ValueBasementSlab
    phx_obj.slab_depth_below_grade_m = _data.DepthBasementBelowGroundSurface
    phx_obj.basement_wall_u_value = _data.U_ValueBasementWall
    return phx_obj


def _PhxUnHeatedBasement(_data: wufi_xml.FoundationInterface) -> PhxUnHeatedBasement:
    phx_obj = PhxUnHeatedBasement()
    phx_obj.display_name = _data.Name
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.slab_depth_below_grade_m = _data.DepthBasementBelowGroundSurface
    phx_obj.basement_wall_height_above_grade_m = _data.HeightBasementWallAboveGrade
    phx_obj.floor_ceiling_area_m2 = _data.FloorSlabArea
    phx_obj.floor_slab_u_value = _data.U_ValueBasementSlab
    phx_obj.floor_ceiling_area_m2 = _data.FloorCeilingArea
    phx_obj.ceiling_u_value = _data.U_ValueCeilingToUnheatedCellar
    phx_obj.basement_wall_uValue_below_grade = _data.U_ValueBasementWall
    phx_obj.basement_wall_uValue_above_grade = _data.U_ValueWallAboveGround
    phx_obj.floor_slab_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.basement_volume_m3 = _data.BasementVolume
    return phx_obj


def _PhxSlabOnGrade(_data: wufi_xml.FoundationInterface) -> PhxSlabOnGrade:
    phx_obj = PhxSlabOnGrade()
    phx_obj.display_name = _data.Name
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.floor_slab_area_m2 = _data.FloorSlabArea
    phx_obj.floor_slab_u_value = _data.U_ValueBasementSlab
    phx_obj.floor_slab_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.perim_insulation_position = _data.PositionPerimeterInsulation
    phx_obj.perim_insulation_width_or_depth_m = _data.PerimeterInsulationWidthDepth
    phx_obj.perim_insulation_thickness_m = _data.ThicknessPerimeterInsulation
    phx_obj.perim_insulation_conductivity = _data.ConductivityPerimeterInsulation
    return phx_obj


def _PhxVentedCrawlspace(_data: wufi_xml.FoundationInterface) -> PhxVentedCrawlspace:
    phx_obj = PhxVentedCrawlspace()
    phx_obj.display_name = _data.Name
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.crawlspace_floor_slab_area_m2 = _data.FloorCeilingArea
    phx_obj.ceiling_above_crawlspace_u_value = _data.U_ValueCeilingToUnheatedCellar
    phx_obj.crawlspace_floor_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.crawlspace_wall_height_above_grade_m = _data.HeightBasementWallAboveGrade
    phx_obj.crawlspace_floor_u_value = _data.U_ValueCrawlspaceFloor
    phx_obj.crawlspace_vent_opening_are_m2 = _data.CrawlspaceVentOpenings
    phx_obj.crawlspace_wall_u_value = _data.U_ValueWallAboveGround
    return phx_obj


def _PhxFoundation(_data: wufi_xml.FoundationInterface) -> PhxFoundation:
    foundation_type_builders = {
        FoundationType.HEATED_BASEMENT: "PhxHeatedBasement",
        FoundationType.UNHEATED_BASEMENT: "PhxUnHeatedBasement",
        FoundationType.SLAB_ON_GRADE: "PhxSlabOnGrade",
        FoundationType.VENTED_CRAWLSPACE: "PhxVentedCrawlspace",
    }

    device_type = FoundationType(_data.FloorSlabType)
    builder_class = foundation_type_builders[device_type]

    # -- Pass the data off to the correct foundation builder class
    return as_phx_obj(_data, builder_class)


# ----------------------------------------------------------------------
# -- Site & Climate


def _PhxPhiusCertification(_data: wufi_xml.PassivehouseData) -> PhxPhiusCertification:
    phx_obj = PhxPhiusCertification()
    phx_obj.use_monthly_shading = _data.UseWUFIMeanMonthShading

    # no idea why PH_BuildingData is a list?....
    bldg_data = _data.PH_Buildings[0]
    phx_obj.ph_building_data = as_phx_obj(bldg_data, "PhxPhBuildingData")

    # ----------------------------------------------------------------------
    # --- Certification Criteria
    criteria = phx_obj.phius_certification_criteria
    criteria.ph_selection_target_data = _data.PH_SelectionTargetData
    criteria.phius_annual_heating_demand = _data.AnnualHeatingDemand
    criteria.phius_annual_cooling_demand = _data.AnnualCoolingDemand
    criteria.phius_peak_heating_load = _data.PeakHeatingLoad
    criteria.phius_peak_cooling_load = _data.PeakCoolingLoad

    # ----------------------------------------------------------------------
    # --- Certification Settings
    settings = phx_obj.phius_certification_settings
    settings.phius_building_certification_program = PhiusCertificationProgram(
        _data.PH_CertificateCriteria
    )
    settings.phius_building_category_type = PhiusCertificationBuildingCategoryType(
        bldg_data.BuildingCategory
    )
    if bldg_data.OccupancyTypeResidential:
        settings.phius_building_use_type = PhiusCertificationBuildingUseType(
            bldg_data.OccupancyTypeResidential
        )
    settings.phius_building_status = PhiusCertificationBuildingStatus(
        bldg_data.BuildingStatus
    )
    settings.phius_building_type = PhiusCertificationBuildingType(bldg_data.BuildingType)

    return phx_obj


def _PhxPhBuildingData(_data: wufi_xml.PH_Building) -> PhxPhBuildingData:
    phx_obj = PhxPhBuildingData()
    phx_obj.id_num = _data.IdentNr
    phx_obj.num_of_units = _data.NumberUnits
    phx_obj.num_of_floors = _data.CountStories
    phx_obj.occupancy_setting_method = _data.OccupancySettingMethod
    phx_obj.airtightness_q50 = _data.EnvelopeAirtightnessCoefficient
    phx_obj.setpoints.winter = _data.IndoorTemperature
    phx_obj.setpoints.summer = _data.OverheatingTemperatureThreshold
    phx_obj.mech_room_temp = _data.MechanicalRoomTemperature
    phx_obj.non_combustible_materials = _data.NonCombustibleMaterials
    phx_obj.summer_hrv_bypass_mode = hvac_enums.PhxSummerBypassMode(
        _data.SummerHRVHumidityRecovery
    )

    for foundation_data in _data.FoundationInterfaces or []:
        phx_obj.add_foundation(as_phx_obj(foundation_data, "PhxFoundation"))

    return phx_obj


def _PhxSite(_data: wufi_xml.ClimateLocation) -> PhxSite:
    phx_obj = PhxSite()
    phx_obj.selection = SiteSelection(_data.Selection)

    ph_climate_data = _data.PH_ClimateLocation
    if ph_climate_data:
        phx_obj.location.latitude = ph_climate_data.Latitude
        phx_obj.location.longitude = ph_climate_data.Longitude
        phx_obj.location.hours_from_UTC = ph_climate_data.dUTC
        phx_obj.location.site_elevation = ph_climate_data.HeightNNBuilding
        phx_obj.location.climate_zone = ph_climate_data.ClimateZone

    phx_obj.climate = as_phx_obj(ph_climate_data, "PhxClimate")
    phx_obj.ground = as_phx_obj(ph_climate_data, "PhxGround")
    phx_obj.energy_factors = as_phx_obj(ph_climate_data, "PhxSiteEnergyFactors")

    return phx_obj


def _PhxClimate(_data: wufi_xml.PH_ClimateLocation) -> PhxClimate:
    def _monthly_values(_in) -> List[float]:
        values = []
        for i in range(12):
            try:
                val = round(getattr(_in[i], "Item", 0.0), 2)
            except IndexError:
                val = 0.0
            values.append(val)
        return values

    phx_obj = PhxClimate()

    phx_obj.station_elevation = _data.HeightNNWeatherStation
    phx_obj.selection = SiteClimateSelection(_data.Selection)
    phx_obj.daily_temp_swing = _data.DailyTemperatureSwingSummer
    phx_obj.avg_wind_speed = _data.AverageWindSpeed

    phx_obj.temperature_air = _monthly_values(_data.TemperatureMonthly)
    phx_obj.temperature_dewpoint = _monthly_values(_data.DewPointTemperatureMonthly)
    phx_obj.temperature_sky = _monthly_values(_data.SkyTemperatureMonthly)
    phx_obj.radiation_north = _monthly_values(_data.NorthSolarRadiationMonthly)
    phx_obj.radiation_east = _monthly_values(_data.EastSolarRadiationMonthly)
    phx_obj.radiation_south = _monthly_values(_data.SouthSolarRadiationMonthly)
    phx_obj.radiation_west = _monthly_values(_data.WestSolarRadiationMonthly)
    phx_obj.radiation_global = _monthly_values(_data.GlobalSolarRadiationMonthly)

    phx_obj.peak_heating_1.temperature_air = _data.TemperatureHeating1
    phx_obj.peak_heating_1.radiation_north = _data.NorthSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_east = _data.EastSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_south = _data.SouthSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_west = _data.WestSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_global = _data.GlobalSolarRadiationHeating1

    phx_obj.peak_heating_2.temperature_air = _data.TemperatureHeating2
    phx_obj.peak_heating_2.radiation_north = _data.NorthSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_east = _data.EastSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_south = _data.SouthSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_west = _data.WestSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_global = _data.GlobalSolarRadiationHeating2

    phx_obj.peak_cooling_1.temperature_air = _data.TemperatureCooling
    phx_obj.peak_cooling_1.radiation_north = _data.NorthSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_east = _data.EastSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_south = _data.SouthSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_west = _data.WestSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_global = _data.GlobalSolarRadiationCooling

    phx_obj.peak_cooling_2.temperature_air = _data.TemperatureCooling2
    phx_obj.peak_cooling_2.radiation_north = _data.NorthSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_east = _data.EastSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_south = _data.SouthSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_west = _data.WestSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_global = _data.GlobalSolarRadiationCooling2

    return phx_obj


def _PhxGround(_data: wufi_xml.PH_ClimateLocation) -> PhxGround:
    phx_obj = PhxGround()
    phx_obj.ground_thermal_conductivity = _data.GroundThermalConductivity
    phx_obj.ground_heat_capacity = _data.GroundHeatCapacitiy
    phx_obj.ground_density = _data.GroundDensity
    phx_obj.depth_groundwater = _data.DepthGroundwater
    phx_obj.flow_rate_groundwater = _data.FlowRateGroundwater
    return phx_obj


def _PhxSiteEnergyFactors(_data: wufi_xml.PH_ClimateLocation) -> PhxSiteEnergyFactors:
    _wufi_order = (
        "OIL",
        "NATURAL_GAS",
        "LPG",
        "HARD_COAL",
        "WOOD",
        "ELECTRICITY_MIX",
        "ELECTRICITY_PV",
        "HARD_COAL_CGS_70_CHP",
        "HARD_COAL_CGS_35_CHP",
        "HARD_COAL_CGS_0_CHP",
        "GAS_CGS_70_CHP",
        "GAS_CGS_35_CHP",
        "GAS_CGS_0_CHP",
        "OIL_CGS_70_CHP",
        "OIL_CGS_35_CHP",
        "OIL_CGS_0_CHP",
    )

    phx_obj = PhxSiteEnergyFactors()
    phx_obj.selection_pe_co2_factor = SiteEnergyFactorSelection(
        _data.SelectionPECO2Factor
    )

    for name, factor in zip(_wufi_order, _data.PEFactorsUserDef):
        phx_obj.pe_factors[name] = PhxPEFactor(factor.__root__, "kWh/kWh", name)

    for name, factor in zip(_wufi_order, _data.CO2FactorsUserDef):
        phx_obj.co2_factors[name] = PhxCO2Factor(factor.__root__, "g/kWh", name)

    return phx_obj


# ----------------------------------------------------------------------
# -- Zones


def _PhxZone(_data: wufi_xml.Zone, _phx_project_host: PhxProject) -> PhxZone:
    def _spec_cap_WH_m2k(_input):
        if _input <= 60.0:
            return 1
        elif _input <= 132.0:
            return 2
        else:
            return 3

    phx_obj = PhxZone()

    phx_obj.id_num = _data.IdentNr
    phx_obj.display_name = _data.Name
    phx_obj.volume_gross = _data.GrossVolume
    phx_obj.volume_net = _data.NetVolume
    phx_obj.weighted_net_floor_area = _data.FloorArea
    phx_obj.clearance_height = _data.ClearanceHeight
    phx_obj.res_occupant_quantity = _data.OccupantQuantityUserDef
    phx_obj.res_number_bedrooms = _data.NumberBedrooms
    phx_obj.specific_heat_capacity = SpecificHeatCapacity(
        _spec_cap_WH_m2k(_data.SpecificHeatCapacity)
    )

    # ----------------------------------------------------------------------
    # -- Create all the spaces from the XML "RoomsVentilation" data
    for space_ventilation_data in _data.RoomsVentilation:
        phx_obj.spaces.append(
            as_phx_obj(
                space_ventilation_data,
                "PhxSpace",
                _phx_project_host=_phx_project_host,
            )
        )

    # -- Try and add in any occupancy, lighting loads to the spaces, if they exist.
    # -- Will try and find a load with a matching name.
    occupancy_load_data = {d.Name: d for d in _data.LoadsPersonsPH or []}
    lighting_load_data = {d.Name: d for d in _data.LoadsLightingsPH or []}
    for space in phx_obj.spaces:
        space = _add_occupancy_data_to_space(
            space, _phx_project_host, occupancy_load_data.get(space.display_name, None)
        )
        space = _add_lighting_data_to_space(
            space, _phx_project_host, lighting_load_data.get(space.display_name, None)
        )

    # ----------------------------------------------------------------------
    # -- Add in any devices and thermal bridges as well.
    for vent_dict_data in _data.ExhaustVents or []:
        new_phx_exhaust = as_phx_obj(vent_dict_data, "PhxExhaustVent")
        phx_obj.exhaust_ventilator_collection.add_new_ventilator(
            new_phx_exhaust.identifier, new_phx_exhaust
        )

    for device_dict_data in _data.HomeDevice or []:
        new_phx_home_device = as_phx_obj(device_dict_data, "PhxHomeDevice")
        phx_obj.elec_equipment_collection.add_new_device(
            new_phx_home_device.identifier, new_phx_home_device
        )

    for i, bridge_dict_data in enumerate(_data.ThermalBridges or []):
        new_phx_tb = as_phx_obj(bridge_dict_data, "PhxComponentThermalBridge")
        new_phx_tb.identifier = f"ThermalBridge_{i :03}"
        phx_obj.add_thermal_bridge(new_phx_tb)

    return phx_obj


def _add_occupancy_data_to_space(
    _space: PhxSpace,
    _phx_project_host: PhxProject,
    _occupancy_data: Optional[wufi_xml.LoadPerson],
) -> PhxSpace:
    """Add occupancy data to a space, if any is supplied."""
    if not _occupancy_data:
        return _space

    project_util_pat_occ = _phx_project_host.utilization_patterns_occupancy
    occ_pattern_id = _occupancy_data.IdentNrUtilizationPattern
    _space.occupancy.schedule = project_util_pat_occ[str(occ_pattern_id)]
    _space.peak_occupancy = _occupancy_data.NumberOccupants

    return _space


def _add_lighting_data_to_space(
    _space: PhxSpace,
    _phx_project_host: PhxProject,
    _lighting_data: Optional[wufi_xml.LoadsLighting],
) -> PhxSpace:
    """Add lighting data to a space, if any is supplied."""
    if not _lighting_data:
        return _space

    # -- Get the right Lighting Utilization Pattern
    project_util_pat_lighting = _phx_project_host.utilization_patterns_lighting
    lighting_pattern_id = _lighting_data.RoomCategory

    # -- Ensure that the lighting pattern is in the project collection. If
    # -- not, create a new one based on the full-load lighting hours.
    if not project_util_pat_lighting.key_is_in_collection(lighting_pattern_id):
        new_schedule = PhxScheduleLighting.from_annual_operating_hours(
            _lighting_data.LightingFullLoadHours
        )
        new_schedule.display_name = _lighting_data.Name
        new_schedule.id_num = int(lighting_pattern_id)
        new_schedule.identifier = str(lighting_pattern_id)
        project_util_pat_lighting.add_new_util_pattern(new_schedule)

    # -- Assign the pattern to the space.
    _space.lighting.schedule = project_util_pat_lighting[str(lighting_pattern_id)]

    # -- Set the lighting load for the space.
    _space.lighting.load.installed_w_per_m2 = _lighting_data.InstalledLightingPower

    return _space


def _PhxSpace(_data: wufi_xml.Room, _phx_project_host: PhxProject) -> PhxSpace:
    phx_obj = PhxSpace()

    # phx_obj.id_num = int(_data."])   phx_obj.display_name = _data.Name
    phx_obj.wufi_type = _data.Type
    phx_obj.quantity = _data.Quantity
    phx_obj.floor_area = _data.AreaRoom
    phx_obj.weighted_floor_area = _data.AreaRoom
    phx_obj.clear_height = _data.ClearRoomHeight

    # -- Ventilation Unit (ERV) number
    phx_obj.vent_unit_id_num = _data.IdentNrVentilationUnit

    # -- Ventilation Schedule and Load
    project_util_pat_ven = _phx_project_host.utilization_patterns_ventilation
    vent_pattern_id = _data.IdentNrUtilizationPatternVent
    phx_obj.ventilation.schedule = project_util_pat_ven[str(vent_pattern_id)]
    phx_obj.ventilation.load.flow_extract = _data.DesignVolumeFlowRateExhaust
    phx_obj.ventilation.load.flow_supply = _data.DesignVolumeFlowRateSupply

    return phx_obj


# ----------------------------------------------------------------------
# -- Mechanicals


def _PhxZoneCoverage(_data: wufi_xml.ZoneCoverage) -> PhxZoneCoverage:
    phx_obj = PhxZoneCoverage()

    phx_obj.zone_num = _data.IdentNrZone
    phx_obj.heating = _data.CoverageHeating
    phx_obj.cooling = _data.CoverageCooling
    phx_obj.ventilation = _data.CoverageVentilation
    phx_obj.humidification = _data.CoverageHumidification
    phx_obj.dehumidification = _data.CoverageDehumidification

    return phx_obj


def _PhxMechanicalDevice(_data: wufi_xml.Device) -> Any:
    device_type_builders = {
        hvac_enums.SystemType.VENTILATION: "PhxDevice_Ventilation",
        hvac_enums.SystemType.ELECTRIC: "PhxDevice_Electric",
        hvac_enums.SystemType.BOILER: "PhxDevice_Boiler",
        hvac_enums.SystemType.DISTRICT_HEAT: "PhxDevice_DistrictHeat",
        hvac_enums.SystemType.HEAT_PUMP: "PhxDevice_HeatPump",
        hvac_enums.SystemType.WATER_STORAGE: "PhxDevice_WaterStorage",
        hvac_enums.SystemType.PHOTOVOLTAIC: "PhxDevice_Photovoltaic",
    }

    device_type = hvac_enums.SystemType(_data.SystemType)
    builder_class = device_type_builders[device_type]

    # -- Pass the data off to the correct device builder class
    return as_phx_obj(_data, builder_class)


def _PhxDevice_Ventilation(_data: wufi_xml.Device) -> PhxDeviceVentilator:
    phx_obj = PhxDeviceVentilator()
    phx_obj.display_name = _data.Name
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)

    phx_obj.params.sensible_heat_recovery = _data.HeatRecovery
    phx_obj.params.latent_heat_recovery = _data.MoistureRecovery
    if _data.PH_Parameters:
        phx_obj.params.quantity = _data.PH_Parameters.Quantity
        phx_obj.params.electric_efficiency = _data.PH_Parameters.ElectricEfficiency
        phx_obj.params.frost_protection_reqd = _data.PH_Parameters.FrostProtection
        phx_obj.params.temperature_below_defrost_used = (
            _data.PH_Parameters.TemperatureBelowDefrostUsed
        )

    return phx_obj


def _PhxDevice_Electric(_data: wufi_xml.Device) -> PhxHeaterElectric:
    return PhxHeaterElectric()


def _PhxDevice_Boiler(_data: wufi_xml.Device) -> Optional[PhxHeaterBoiler]:
    boiler_builders = {
        hvac_enums.PhxFuelType.NATURAL_GAS: "PhxHeaterBoilerFossil",
        hvac_enums.PhxFuelType.OIL: "PhxHeaterBoilerFossil",
        hvac_enums.PhxFuelType.WOOD_LOG: "PhxHeaterBoilerWood",
        hvac_enums.PhxFuelType.WOOD_PELLET: "PhxHeaterBoilerWood",
    }
    if not _data.PH_Parameters:
        return None

    boiler_type = hvac_enums.PhxFuelType(_data.PH_Parameters.EnergySourceBoilerType)
    return as_phx_obj(_data, boiler_builders[boiler_type])


def _PhxHeaterBoilerFossil(_data: wufi_xml.Device) -> PhxHeaterBoilerFossil:
    phx_obj = PhxHeaterBoilerFossil()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.params.fuel = _data.PH_Parameters.EnergySourceBoilerType
    phx_obj.params.condensing = _data.PH_Parameters.CondensingBoiler
    phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace
    phx_obj.params.effic_at_30_percent_load = _data.PH_Parameters.BoilerEfficiency30
    phx_obj.params.effic_at_nominal_load = (
        _data.PH_Parameters.BoilerEfficiencyNominalOutput
    )
    phx_obj.params.avg_rtrn_temp_at_30_percent_load = (
        _data.PH_Parameters.AverageReturnTemperatureMeasured30Load
    )
    phx_obj.params.avg_temp_at_70C_55C = (
        _data.PH_Parameters.AverageBoilerTemperatureDesign70_55
    )
    phx_obj.params.avg_temp_at_55C_45C = (
        _data.PH_Parameters.AverageBoilerTemperatureDesign55_45
    )
    phx_obj.params.avg_temp_at_32C_28C = (
        _data.PH_Parameters.AverageBoilerTemperatureDesign35_28
    )
    phx_obj.params.standby_loss_at_70C = _data.PH_Parameters.StandbyHeatLossBoiler70
    phx_obj.params.rated_capacity = _data.PH_Parameters.MaximalBoilerPower
    return phx_obj


def _PhxHeaterBoilerWood(_data: wufi_xml.Device) -> PhxHeaterBoilerWood:
    phx_obj = PhxHeaterBoilerWood()

    return phx_obj


def _PhxDevice_DistrictHeat(_data: wufi_xml.Device) -> PhxHeaterDistrictHeat:
    return PhxHeaterDistrictHeat()


def _PhxDevice_HeatPump(_data: wufi_xml.Device) -> Optional[PhxHeaterHeatPump]:
    hp_builders = {
        hvac_enums.HeatPumpType.COMBINED: "PhxDevice_HeatPump_Combined",
        hvac_enums.HeatPumpType.ANNUAL: "PhxDevice_HeatPump_Annual",
        hvac_enums.HeatPumpType.RATED_MONTHLY: "PhxDevice_HeatPump_RatedMonthly",
        hvac_enums.HeatPumpType.HOT_WATER: "PhxDevice_HeatPump_HotWater",
    }
    if _data.PH_Parameters and _data.PH_Parameters.HPType:
        hp_type = hvac_enums.HeatPumpType(_data.PH_Parameters.HPType)
        builder_class = hp_builders[hp_type]
        return as_phx_obj(_data, builder_class)
    else:
        return None


def _PhxDevice_HeatPump_Combined(_data: wufi_xml.Device) -> PhxHeaterHeatPumpCombined:
    phx_obj = PhxHeaterHeatPumpCombined()
    return phx_obj


def _PhxDevice_HeatPump_Annual(_data: wufi_xml.Device) -> PhxHeaterHeatPumpAnnual:
    phx_obj = PhxHeaterHeatPumpAnnual()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.params.annual_COP = _data.PH_Parameters.AnnualCOP
    phx_obj.params.total_system_perf_ratio = (
        _data.PH_Parameters.TotalSystemPerformanceRatioHeatGenerator
    )
    return phx_obj


def _PhxDevice_HeatPump_RatedMonthly(_data: wufi_xml.Device) -> PhxHeaterHeatPumpMonthly:
    phx_obj = PhxHeaterHeatPumpMonthly()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.params.COP_1 = _data.PH_Parameters.RatedCOP1
    phx_obj.params.COP_2 = _data.PH_Parameters.RatedCOP2
    phx_obj.params.ambient_temp_1 = _data.PH_Parameters.AmbientTemperature1
    phx_obj.params.ambient_temp_2 = _data.PH_Parameters.AmbientTemperature2
    return phx_obj


def _PhxDevice_HeatPump_HotWater(_data: wufi_xml.Device) -> PhxHeaterHeatPumpHotWater:
    phx_obj = PhxHeaterHeatPumpHotWater()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.params.annual_COP = _data.PH_Parameters.AnnualCOP
    phx_obj.params.annual_system_perf_ratio = (
        _data.PH_Parameters.TotalSystemPerformanceRatioHeatGenerator
    )
    phx_obj.params.annual_energy_factor = _data.PH_Parameters.HPWH_EF
    return phx_obj


def _PhxDevice_WaterStorage(_data: wufi_xml.Device) -> PhxHotWaterTank:
    phx_obj = PhxHotWaterTank()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.quantity = _data.PH_Parameters.QauntityWS
    phx_obj.params.display_name = _data.Name
    phx_obj.params.input_option = hvac_enums.PhxHotWaterInputOptions(
        _data.PH_Parameters.InputOption
    )
    phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace
    phx_obj.params.storage_loss_rate = _data.PH_Parameters.AverageHeatReleaseStorage

    phx_obj.params.storage_capacity = _data.PH_Parameters.SolarThermalStorageCapacity
    phx_obj.params.standby_losses = _data.PH_Parameters.StorageLossesStandby
    phx_obj.params.room_temp = _data.PH_Parameters.TankRoomTemp
    phx_obj.params.water_temp = _data.PH_Parameters.TypicalStorageWaterTemperature
    return phx_obj


def _PhxDevice_Photovoltaic(_data: wufi_xml.Device) -> PhxDevicePhotovoltaic:
    phx_obj = PhxDevicePhotovoltaic()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.params.location_type = _data.PH_Parameters.SelectionLocation
    phx_obj.params.onsite_utilization_type = (
        _data.PH_Parameters.SelectionOnSiteUtilization
    )

    phx_obj.params.utilization_type = _data.PH_Parameters.SelectionUtilization
    phx_obj.params.array_size = _data.PH_Parameters.ArraySizePV
    phx_obj.params.photovoltaic_renewable_energy = (
        _data.PH_Parameters.PhotovoltaicRenewableEnergy
    )

    phx_obj.params.onsite_utilization_factor = _data.PH_Parameters.OnsiteUtilization
    phx_obj.params.auxiliary_energy = _data.PH_Parameters.AuxiliaryEnergy
    phx_obj.params.auxiliary_energy_DHW = _data.PH_Parameters.AuxiliaryEnergyDHW
    phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace

    return phx_obj


# ----------------------------------------------------------------------
# -- Exhaust Ventilators


def _PhxExhaustVent(_data: wufi_xml.ExhaustVent) -> AnyPhxExhaustVent:
    exhaust_vent_builders = {
        hvac_enums.PhxExhaustVentType.KITCHEN_HOOD: "PhxExhaustVent_KitchenHood",
        hvac_enums.PhxExhaustVentType.DRYER: "PhxExhaustVent_Dryer",
        hvac_enums.PhxExhaustVentType.USER_DEFINED: "PhxExhaustVent_UserDefined",
    }
    vent_type = hvac_enums.PhxExhaustVentType(_data.Type)
    return as_phx_obj(_data, exhaust_vent_builders[vent_type])


def _PhxExhaustVent_KitchenHood(
    _data: wufi_xml.ExhaustVent,
) -> PhxExhaustVentilatorRangeHood:
    phx_obj = PhxExhaustVentilatorRangeHood()
    phx_obj.params.annual_runtime_minutes = _data.RunTimePerYear
    phx_obj.params.exhaust_flow_rate_m3h = _data.ExhaustVolumeFlowRate
    return phx_obj


def _PhxExhaustVent_Dryer(_data: wufi_xml.ExhaustVent) -> PhxExhaustVentilatorDryer:
    phx_obj = PhxExhaustVentilatorDryer()
    phx_obj.params.annual_runtime_minutes = _data.RunTimePerYear
    phx_obj.params.exhaust_flow_rate_m3h = _data.ExhaustVolumeFlowRate
    return phx_obj


def _PhxExhaustVent_UserDefined(
    _data: wufi_xml.ExhaustVent,
) -> PhxExhaustVentilatorUserDefined:
    phx_obj = PhxExhaustVentilatorUserDefined()
    phx_obj.params.annual_runtime_minutes = _data.RunTimePerYear
    phx_obj.params.exhaust_flow_rate_m3h = _data.ExhaustVolumeFlowRate
    return phx_obj


# ----------------------------------------------------------------------
# -- Electrical Devices


def _PhxHomeDevice(_data: wufi_xml.H_Device) -> PhxElectricalDevice:
    device_builders = {
        ElectricEquipmentType.DISHWASHER: "PhxDeviceDishwasher",
        ElectricEquipmentType.CLOTHES_WASHER: "PhxDeviceClothesWasher",
        ElectricEquipmentType.CLOTHES_DRYER: "PhxDeviceClothesDryer",
        ElectricEquipmentType.REFRIGERATOR: "PhxDeviceRefrigerator",
        ElectricEquipmentType.FREEZER: "PhxDeviceFreezer",
        ElectricEquipmentType.FRIDGE_FREEZER: "PhxDeviceFridgeFreezer",
        ElectricEquipmentType.COOKING: "PhxDeviceCooktop",
        ElectricEquipmentType.CUSTOM: "PhxDeviceCustomElec",
        ElectricEquipmentType.MEL: "PhxDeviceMEL",
        ElectricEquipmentType.LIGHTING_INTERIOR: "PhxDeviceLightingInterior",
        ElectricEquipmentType.LIGHTING_EXTERIOR: "PhxDeviceLightingExterior",
        ElectricEquipmentType.LIGHTING_GARAGE: "PhxDeviceLightingGarage",
        ElectricEquipmentType.CUSTOM_LIGHTING: "PhxDeviceCustomLighting",
        ElectricEquipmentType.CUSTOM_MEL: "PhxDeviceCustomMEL",
    }
    device_type = ElectricEquipmentType(_data.Type)
    phx_obj: PhxElectricalDevice = as_phx_obj(_data, device_builders[device_type])

    # -- Set all the common parameters
    phx_obj.comment = _data.Comment
    phx_obj.reference_quantity = _data.ReferenceQuantity
    phx_obj.quantity = _data.Quantity
    phx_obj.in_conditioned_space = _data.InConditionedSpace
    phx_obj.reference_energy_norm = _data.ReferenceEnergyDemandNorm
    phx_obj.energy_demand = _data.EnergyDemandNorm
    phx_obj.energy_demand_per_use = _data.EnergyDemandNormUse
    phx_obj.combined_energy_factor = _data.CEF_CombinedEnergyFactor

    return phx_obj


def _PhxDeviceDishwasher(_data: wufi_xml.H_Device) -> PhxDeviceDishwasher:
    phx_obj = PhxDeviceDishwasher()
    phx_obj.water_connection = _data.Connection
    phx_obj.capacity_type = _data.DishwasherCapacityPreselection
    phx_obj.capacity = _data.DishwasherCapacityInPlace
    return phx_obj


def _PhxDeviceClothesWasher(_data: wufi_xml.H_Device) -> PhxDeviceClothesWasher:
    phx_obj = PhxDeviceClothesWasher()
    phx_obj.water_connection = _data.Connection
    phx_obj.utilization_factor = _data.UtilizationFactor
    phx_obj.capacity = _data.CapacityClothesWasher
    phx_obj.modified_energy_factor = _data.MEF_ModifiedEnergyFactor
    return phx_obj


def _PhxDeviceClothesDryer(_data: wufi_xml.H_Device) -> PhxDeviceClothesDryer:
    phx_obj = PhxDeviceClothesDryer()
    phx_obj.dryer_type = _data.Dryer_Choice or 4
    phx_obj.gas_consumption = _data.GasConsumption
    phx_obj.gas_efficiency_factor = _data.EfficiencyFactorGas
    phx_obj.field_utilization_factor_type = _data.FieldUtilizationFactorPreselection
    phx_obj.field_utilization_factor = _data.FieldUtilizationFactor
    return phx_obj


def _PhxDeviceRefrigerator(_data: wufi_xml.H_Device) -> PhxDeviceRefrigerator:
    phx_obj = PhxDeviceRefrigerator()
    return phx_obj


def _PhxDeviceFreezer(_data: wufi_xml.H_Device) -> PhxDeviceFreezer:
    phx_obj = PhxDeviceFreezer()
    return phx_obj


def _PhxDeviceFridgeFreezer(_data: wufi_xml.H_Device) -> PhxDeviceFridgeFreezer:
    phx_obj = PhxDeviceFridgeFreezer()
    return phx_obj


def _PhxDeviceCooktop(_data: wufi_xml.H_Device) -> PhxDeviceCooktop:
    phx_obj = PhxDeviceCooktop()
    phx_obj.cooktop_type = _data.CookingWith
    return phx_obj


def _PhxDeviceCustomElec(_data: wufi_xml.H_Device) -> PhxDeviceCustomElec:
    phx_obj = PhxDeviceCustomElec()
    return phx_obj


def _PhxDeviceMEL(_data: wufi_xml.H_Device) -> PhxDeviceMEL:
    phx_obj = PhxDeviceMEL()
    return phx_obj


def _PhxDeviceLightingInterior(_data: wufi_xml.H_Device) -> PhxDeviceLightingInterior:
    phx_obj = PhxDeviceLightingInterior()
    phx_obj.frac_high_efficiency = _data.FractionHightEfficiency
    return phx_obj


def _PhxDeviceLightingExterior(_data: wufi_xml.H_Device) -> PhxDeviceLightingExterior:
    phx_obj = PhxDeviceLightingExterior()
    phx_obj.frac_high_efficiency = _data.FractionHightEfficiency
    return phx_obj


def _PhxDeviceLightingGarage(_data: wufi_xml.H_Device) -> PhxDeviceLightingGarage:
    phx_obj = PhxDeviceLightingGarage()
    phx_obj.frac_high_efficiency = _data.FractionHightEfficiency
    return phx_obj


def _PhxDeviceCustomLighting(_data: wufi_xml.H_Device) -> PhxDeviceCustomLighting:
    phx_obj = PhxDeviceCustomLighting()

    return phx_obj


def _PhxDeviceCustomMEL(_data: wufi_xml.H_Device) -> PhxDeviceCustomMEL:
    phx_obj = PhxDeviceCustomMEL()

    return phx_obj



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Water Piping Distribution Objects."""

from __future__ import annotations
from typing import Dict, List, Any, Optional, Tuple
from dataclasses import dataclass, field
import math

from ladybug_geometry.geometry3d.polyline import LineSegment3D


@dataclass
class PhxRecirculationParameters:
    calc_method = 1  # Simplified individual pipes
    pipe_material = 1  # Copper M
    demand_recirc = True
    num_bathrooms = 1
    hot_water_fixtures = 1
    all_pipes_insulated = True
    units_or_floors = 1
    pipe_diameter_m = 1.0  # 3/8" Copper
    air_temp = 20.0  # Deg C
    water_temp = 60.0  # Deg C
    daily_recirc_hours = 24


@dataclass
class PhxPipeSegment:
    """An individual Pipe Segment."""

    identifier: str
    display_name: str
    geometry: LineSegment3D
    diameter_m: float
    insulation_thickness_m: float
    insulation_conductivity: float  # W/mk
    insulation_reflective: bool
    insulation_quality: Any
    daily_period: float

    pipe_wall_thickness_m: float = 0.00225

    @property
    def diameter_mm(self):
        return self.diameter_m * 1000

    @property
    def diameter_inner_m(self) -> float:
        return self.diameter_m

    @property
    def diameter_outer_m(self) -> float:
        """Return the outside diameter including the pipe wall thickness."""
        return self.diameter_inner_m + self.pipe_wall_thickness_m

    @property
    def diameter_with_insulation_m(self) -> float:
        return self.diameter_outer_m + (2 * self.insulation_thickness_m)

    @property
    def length_m(self) -> float:
        return self.geometry.length

    @property
    def pipe_heat_loss_coefficient(self) -> float:
        """Return the pipe's heat-loss-coefficient (W/mk) considering the diameter and insulation."""
        return self._solve_for_pipe_heat_loss_coeff()

    @property
    def _starting_alpha(self) -> float:
        """Return W/m2k based on insulation reflectivity."""
        if self.insulation_reflective:
            return 5.0
        else:
            return 8.0

    def _calc_pipe_heat_loss_coeff(self, _alpha) -> float:
        """Calculate the pipe heat-loss coefficient (W/mk) with a known Alpha (W/m2k) value."""
        _a = self.diameter_outer_m / self.diameter_inner_m
        _b = self.diameter_with_insulation_m / self.diameter_outer_m
        _c = (
            1 / 2 / 55 * math.log(_a)
            + 1 / 2 / self.insulation_conductivity * math.log(_b)
            + 1 / self.diameter_with_insulation_m / _alpha
        )
        return math.pi / _c

    def _calc_pipe_surface_temp(self, dT, _k) -> float:
        """Return a surface temp (k) for the pipe with a known heat-loss-coefficient (W/mk) value."""
        _a = self.diameter_outer_m / self.diameter_inner_m
        _b = self.diameter_with_insulation_m / self.diameter_outer_m
        _c = 1 / 2 / 55 * math.log(_a) + 1 / 2 / self.insulation_conductivity * math.log(
            _b
        )
        return dT - 1 / math.pi * _c * _k * dT

    def _solve_for_pipe_heat_loss_coeff(
        self,
        _alpha: Optional[float] = None,
        _k1: float = 100.0,
        _k2: float = 1.0,
        _surface_temp: Optional[float] = None,
    ) -> float:
        """Return a heat-loss coefficient (W/mk) considering the diameter and insulation.

        Approximates the recursive algorithm found in the PHPP "DHW" worksheet.
        """
        TOLERANCE = 0.001
        DELTA_T = 30  # K

        while _k1 - _k2 > TOLERANCE:
            # -- hang onto the old k to test against tolerance later
            _k1 = _k2

            # -- If W/m2k alpha is not provided (ie: at the start) use the starting-alpha constant
            if not _alpha:
                _alpha = self._starting_alpha

            # -- Calc the new K (W/mk), and new surface temp of the pipe
            _k2 = self._calc_pipe_heat_loss_coeff(_alpha)
            _surface_temp = self._calc_pipe_surface_temp(DELTA_T, _k2)

            # -- calc new alpha W/m2k value
            if self.insulation_reflective:
                _n = 0.1
            else:
                _n = 0.85
            _alpha = _n * 4.8 + 1.62 * _surface_temp**0.333

            # -- Run the solver until the k result is < the tolerance
            self._solve_for_pipe_heat_loss_coeff(_alpha, _k1, _k2, _surface_temp)

        return _k2


@dataclass
class PhxPipeElement:
    """A Pipe Element / Run made of one or more PhxPipeSegments."""

    identifier: str
    display_name: str
    _segments: Dict = field(default_factory=dict)

    @property
    def segments(self) -> List[PhxPipeSegment]:
        return list(self._segments.values())

    @property
    def length_m(self) -> float:
        """Return the total length of all the Pipe Segments."""
        return sum(_.length_m for _ in self.segments) or 1.0

    @property
    def weighted_pipe_heat_loss_coefficient(self):
        """Return a length-weighted total heat loss coefficient (W/mk)"""
        weighted_total = 0.0
        for segment in self.segments:
            weighted_total += segment.pipe_heat_loss_coefficient * segment.length_m
        return weighted_total / self.length_m

    @property
    def weighted_diameter_mm(self):
        """Return a length-weighted total diameter (mm)"""
        weighted_total = 0.0
        for segment in self.segments:
            weighted_total += segment.diameter_mm * segment.length_m
        return weighted_total / self.length_m

    def add_segment(self, _s: PhxPipeSegment) -> None:
        self._segments[_s.identifier] = _s



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Space (Room) Class"""

from __future__ import annotations
from typing import ClassVar
from dataclasses import dataclass, field

from PHX.model.programs import occupancy as occupancy_prog
from PHX.model.programs.ventilation import PhxProgramVentilation
from PHX.model.programs.lighting import PhxProgramLighting


@dataclass
class PhxSpace:
    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    display_name: str = "Unnamed_Space"
    wufi_type: int = 99  # User Determined
    quantity: int = 1
    floor_area: float = 0.0
    weighted_floor_area: float = 0.0
    net_volume: float = 0.0
    clear_height: float = 2.5

    # -- Ventilation Unit (ERV) number
    vent_unit_id_num: int = 0

    # -- Programs
    ventilation: PhxProgramVentilation = field(default_factory=PhxProgramVentilation)
    occupancy: occupancy_prog.PhxProgramOccupancy = field(
        default_factory=occupancy_prog.PhxProgramOccupancy
    )
    lighting: PhxProgramLighting = field(default_factory=PhxProgramLighting)

    electric_equipment = None

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def peak_occupancy(self) -> float:
        """Returns the peak occupancy for the space (Num. of people)."""
        return self.occupancy.load.people_per_m2 * self.floor_area

    @peak_occupancy.setter
    def peak_occupancy(self, value: float) -> None:
        """Sets the peak occupancy for the space (Num. of people)."""
        self.occupancy.load.people_per_m2 = value / self.floor_area



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Pydantic Model for WUFI-XML file format."""

from __future__ import annotations
from typing import Optional, List, Dict, Union, Any
from pydantic import BaseModel, validator
from PHX.from_WUFI_XML.read_WUFI_XML_file import Tag
from ph_units.converter import convert
from PHX.from_WUFI_XML import wufi_file_types as unit

# ------------------------------------------------------------------------------
# -- Unit Types


def unpack_xml_tag(_value: Union[List, Dict, Tag]) -> Union[Dict, List, str, None]:
    """This validator should run first before any unit-conversion. This will
    unpack the Tag object and return the .text, or a dict with the unit and the
    value if it is a unit-type.
    """
    if isinstance(_value, Tag):
        # -- If it is a unit-type, pass back a dict with the unit and the value
        # -- this will get unpacked and used by the unit-type converter.
        unit_type = getattr(_value, "attrib", {}).get("unit", None)
        if unit_type:
            return {"value": _value.text, "unit_type": unit_type}

        # -- Otherwise, just pass back the .text
        if _value.text == "None":
            return None
        else:
            return _value.text

    return _value


# ------------------------------------------------------------------------------
# -- Climate


class CO2FactorsUserDef(BaseModel):
    __root__: Any

    @validator("__root__", pre=True)
    def unpack_xml_tag_name(cls, v):
        # -- Have to do it this way cus' the WUFI file structure is a mess
        # Input is a dict like: '{'PEF0': Tag(text='1.1', tag='PEF0', attrib={'unit': 'Btu/Btu'})}'
        input_tag = list(v.values())[-1]
        input_dict: Dict = unpack_xml_tag(input_tag)  # type: ignore
        result = convert(input_dict["value"], str(input_dict["unit_type"]), "G/KWH")
        if not result:
            msg = f"Error converting {input_dict['value']} from {input_dict['unit_type']} to G/KWH"
            raise Exception(msg)
        return float(result)


class PEFactorsUserDef(BaseModel):
    __root__: Any

    @validator("__root__", pre=True)
    def unpack_xml_tag_name(cls, v):
        # -- Have to do it this way cus' the WUFI file structure is a mess
        # Input is a dict like: '{'PEF0': Tag(text='1.1', tag='PEF0', attrib={'unit': 'Btu/Btu'})}'
        input_tag = list(v.values())[-1]
        input_dict: Dict = unpack_xml_tag(input_tag)  # type: ignore
        result = convert(input_dict["value"], str(input_dict["unit_type"]), "KWH/KWH")
        if not result:
            msg = f"Error converting {input_dict['value']} from {input_dict['unit_type']} to KWH/KWH"
            raise Exception(msg)
        return float(result)


class MonthlyClimateTemp_Item(BaseModel):
    Item: Optional[unit.DegreeC]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class MonthlyClimateRadiation_Item(BaseModel):
    Item: Optional[unit.kWh_per_M2]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class PH_ClimateLocation(BaseModel):
    Selection: int
    Name: Optional[str]
    Comment: Optional[str]
    DailyTemperatureSwingSummer: unit.DegreeDeltaK
    AverageWindSpeed: unit.M_per_Second

    Latitude: unit.CardinalDegrees
    Longitude: unit.CardinalDegrees
    dUTC: int

    HeightNNWeatherStation: unit.M
    HeightNNBuilding: unit.M

    ClimateZone: int
    GroundThermalConductivity: unit.Watts_per_MK
    GroundHeatCapacitiy: unit.Joule_per_KGK
    GroundDensity: unit.KG_per_M3
    DepthGroundwater: unit.M
    FlowRateGroundwater: unit.M_per_Day

    TemperatureMonthly: List[Optional[MonthlyClimateTemp_Item]]
    DewPointTemperatureMonthly: List[Optional[MonthlyClimateTemp_Item]]
    SkyTemperatureMonthly: List[Optional[MonthlyClimateTemp_Item]]

    NorthSolarRadiationMonthly: List[Optional[MonthlyClimateRadiation_Item]]
    EastSolarRadiationMonthly: List[Optional[MonthlyClimateRadiation_Item]]
    SouthSolarRadiationMonthly: List[Optional[MonthlyClimateRadiation_Item]]
    WestSolarRadiationMonthly: List[Optional[MonthlyClimateRadiation_Item]]
    GlobalSolarRadiationMonthly: List[Optional[MonthlyClimateRadiation_Item]]

    TemperatureHeating1: Optional[unit.DegreeC]
    NorthSolarRadiationHeating1: Optional[unit.Watts_per_M2]
    EastSolarRadiationHeating1: Optional[unit.Watts_per_M2]
    SouthSolarRadiationHeating1: Optional[unit.Watts_per_M2]
    WestSolarRadiationHeating1: Optional[unit.Watts_per_M2]
    GlobalSolarRadiationHeating1: Optional[unit.Watts_per_M2]

    TemperatureHeating2: Optional[unit.DegreeC]
    NorthSolarRadiationHeating2: Optional[unit.Watts_per_M2]
    EastSolarRadiationHeating2: Optional[unit.Watts_per_M2]
    SouthSolarRadiationHeating2: Optional[unit.Watts_per_M2]
    WestSolarRadiationHeating2: Optional[unit.Watts_per_M2]
    GlobalSolarRadiationHeating2: Optional[unit.Watts_per_M2]

    TemperatureCooling: Optional[unit.DegreeC]
    NorthSolarRadiationCooling: Optional[unit.Watts_per_M2]
    EastSolarRadiationCooling: Optional[unit.Watts_per_M2]
    SouthSolarRadiationCooling: Optional[unit.Watts_per_M2]
    WestSolarRadiationCooling: Optional[unit.Watts_per_M2]
    GlobalSolarRadiationCooling: Optional[unit.Watts_per_M2]

    TemperatureCooling2: Optional[unit.DegreeC]
    NorthSolarRadiationCooling2: Optional[unit.Watts_per_M2]
    EastSolarRadiationCooling2: Optional[unit.Watts_per_M2]
    SouthSolarRadiationCooling2: Optional[unit.Watts_per_M2]
    WestSolarRadiationCooling2: Optional[unit.Watts_per_M2]
    GlobalSolarRadiationCooling2: Optional[unit.Watts_per_M2]

    SelectionPECO2Factor: int
    PEFactorsUserDef: List[PEFactorsUserDef]
    CO2FactorsUserDef: List[CO2FactorsUserDef]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class ClimateLocation(BaseModel):
    Selection: int

    Latitude_DB: Optional[unit.CardinalDegrees]
    Longitude_DB: Optional[unit.CardinalDegrees]
    HeightNN_DB: Optional[unit.M]
    dUTC_DB: Optional[float]

    Albedo: int
    GroundReflShort: unit._Percentage
    GroundReflLong: unit._Percentage
    GroundEmission: unit._Percentage
    CloudIndex: unit._Percentage
    CO2concenration: unit.MG_per_M3
    Unit_CO2concentration: unit.PartsPerMillionByVolume
    PH_ClimateLocation: Optional[PH_ClimateLocation]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Geometry


class Vertix(BaseModel):
    # Note: Bizarrely, even in IP units WUFI XML, the Vertices are in Meters.
    IdentNr: int
    X: float
    Y: float
    Z: float

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class IdentNr(BaseModel):
    IdentNr: int
    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Polygon(BaseModel):
    IdentNr: int
    NormalVectorX: float
    NormalVectorY: float
    NormalVectorZ: float
    IdentNrPoints: List[IdentNr]
    IdentNrPolygonsInside: Optional[List[IdentNr]]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Graphics_3D(BaseModel):
    Vertices: List[Vertix]
    Polygons: List[Polygon]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Building and Components


class Room(BaseModel):
    Name: str
    Type: int
    IdentNrUtilizationPatternVent: int
    IdentNrVentilationUnit: int
    Quantity: int
    AreaRoom: unit.M2
    ClearRoomHeight: unit.M
    DesignVolumeFlowRateSupply: unit.M3_per_Hour
    DesignVolumeFlowRateExhaust: unit.M3_per_Hour

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class LoadPerson(BaseModel):
    Name: str
    IdentNrUtilizationPattern: int
    ChoiceActivityPersons: int
    NumberOccupants: unit._Float
    FloorAreaUtilizationZone: unit.M2

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class LoadsLighting(BaseModel):
    Name: str
    RoomCategory: int
    ChoiceLightTransmissionGlazing: int
    LightingControl: int
    WithinThermalEnvelope: bool
    MotionDetector: bool
    FacadeIncludingWindows: bool
    FractionTreatedFloorArea: unit._Percentage
    DeviationFromNorth: unit.AngleDegree
    RoomDepth: unit.M
    RoomWidth: unit.M
    RoomHeight: unit.M
    LintelHeight: unit.M
    WindowWidth: unit.M
    InstalledLightingPower: unit.Watts_per_M2
    LightingFullLoadHours: unit.Hours_per_Year

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class ExhaustVent(BaseModel):
    Name: Optional[str]
    Type: int
    ExhaustVolumeFlowRate: Optional[unit.M3_per_Hour]
    RunTimePerYear: Optional[unit.Hours_per_Year]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class H_Device(BaseModel):
    # -- Basic
    Comment: Optional[str]
    ReferenceQuantity: Optional[int]
    Quantity: Optional[int]
    InConditionedSpace: Optional[bool]
    ReferenceEnergyDemandNorm: Optional[int]
    EnergyDemandNorm: Optional[unit.kWh]
    EnergyDemandNormUse: Optional[unit.kWh]
    CEF_CombinedEnergyFactor: Optional[unit._Percentage]
    Type: int

    # -- Cooktops
    CookingWith: Optional[int]

    # -- Lighting
    FractionHightEfficiency: Optional[unit._Percentage]

    # -- Dishwasher
    CookingWith: Optional[int]

    # -- Dryer
    Dryer_Choice: Optional[int]
    GasConsumption: Optional[unit.kWh]
    EfficiencyFactorGas: Optional[unit._Percentage]
    FieldUtilizationFactorPreselection: Optional[int]
    FieldUtilizationFactor: Optional[unit._Float]

    #  -- Washer
    Connection: Optional[int]
    UtilizationFactor: Optional[unit._Float]
    CapacityClothesWasher: Optional[unit.M3]
    MEF_ModifiedEnergyFactor: Optional[unit._Float]

    # -- Dishwasher
    DishwasherCapacityPreselection: Optional[int]
    DishwasherCapacityInPlace: Optional[unit._Float]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class IdentNrPolygons(BaseModel):
    IdentNr: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Component(BaseModel):
    IdentNr: int
    Name: str
    Visual: bool
    Type: int
    IdentNrColorI: int
    IdentNrColorE: int
    InnerAttachment: int
    OuterAttachment: int
    IdentNr_ComponentInnerSurface: Optional[int]
    IdentNrAssembly: Optional[int]
    IdentNrWindowType: Optional[int]
    IdentNrPolygons: List[IdentNrPolygons]

    # Window-Specific Attributes
    DepthWindowReveal: Optional[unit.M]
    IdentNrSolarProtection: Optional[int]
    IdentNrOverhang: Optional[int]
    DefaultCorrectionShadingMonth: Optional[unit._Percentage]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class ThermalBridge(BaseModel):
    Name: Optional[str]
    Type: int
    Length: Optional[unit.M]
    PsiValue: Optional[unit.Watts_per_MK]
    IdentNrOptionalClimate: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Zone(BaseModel):
    Name: str
    KindZone: int
    IdentNr: int
    RoomsVentilation: List[Room]
    LoadsPersonsPH: Optional[List[LoadPerson]]
    LoadsLightingsPH: Optional[List[LoadsLighting]]
    GrossVolume_Selection: int
    GrossVolume: unit.M3
    NetVolume_Selection: int
    NetVolume: unit.M3
    FloorArea_Selection: int
    FloorArea: unit.M2
    ClearanceHeight_Selection: int
    ClearanceHeight: unit.M
    SpecificHeatCapacity_Selection: int
    SpecificHeatCapacity: unit.Wh_per_M2K
    IdentNrPH_Building: int
    OccupantQuantityUserDef: unit._Int
    NumberBedrooms: unit._Int
    SummerNaturalVentilationDay: unit.ACH
    SummerNaturalVentilationNight: unit.ACH

    HomeDevice: Optional[List[H_Device]]
    ExhaustVents: Optional[List[ExhaustVent]]
    ThermalBridges: Optional[List[ThermalBridge]]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Building(BaseModel):
    Components: List[Component]
    Zones: List[Zone]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- HVAC


class DistributionDHW(BaseModel):
    CalculationMethodIndividualPipes: int
    DemandRecirculation: bool
    SelectionhotWaterFixtureEff: int
    NumberOfBathrooms: int
    AllPipesAreInsulated: bool
    SelectionUnitsOrFloors: int
    PipeMaterialSimplifiedMethod: int
    PipeDiameterSimplifiedMethod: Optional[float]
    TemperatureRoom_WR: Optional[unit.DegreeC]
    DesignFlowTemperature_WR: Optional[unit.DegreeC]
    DailyRunningHoursCirculation_WR: Optional[unit.Hour]
    LengthCirculationPipes_WR: Optional[unit.M]
    HeatLossCoefficient_WR: Optional[unit.Watts_per_MK]
    LengthIndividualPipes_WR: Optional[unit.M]
    ExteriorPipeDiameter_WR: Optional[unit.M]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class SupportiveDevice(BaseModel):
    Name: Optional[str]
    Type: int
    Quantity: int
    InConditionedSpace: Optional[bool]
    NormEnergyDemand: Optional[unit.Watts]
    Controlled: Optional[bool]
    PeriodOperation: Optional[float]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class PHDistribution(BaseModel):
    DistributionDHW: Optional[DistributionDHW]
    # TODO DistributionHeating: DistributionHeating
    # TODO DistributionCooling: DistributionCooling
    # TODO DistributionVentilation: List[Duct]
    UseDefaultValues: bool
    DeviceInConditionedSpace: Optional[bool]
    SupportiveDevices: Optional[List[SupportiveDevice]]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class ZoneCoverage(BaseModel):
    IdentNrZone: int
    CoverageHeating: float
    CoverageCooling: float
    CoverageVentilation: float
    CoverageHumidification: float
    CoverageDehumidification: float

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class PH_Parameters(BaseModel):
    # -- Heat Pump
    InConditionedSpace: Optional[bool]
    AuxiliaryEnergy: Optional[unit.Watts]
    AuxiliaryEnergyDHW: Optional[unit.Watts]
    AnnualCOP: Optional[unit.kWh_per_kWh]
    TotalSystemPerformanceRatioHeatGenerator: Optional[unit._Percentage]
    HPType: Optional[int]
    RatedCOP1: Optional[float]
    RatedCOP2: Optional[float]
    AmbientTemperature1: Optional[unit.DegreeC]
    AmbientTemperature2: Optional[unit.DegreeC]

    # -- Ventilation
    Quantity: Optional[int]
    HumidityRecoveryEfficiency: Optional[unit._Percentage]
    ElectricEfficiency: Optional[unit.Wh_per_M3]
    DefrostRequired: Optional[bool]
    FrostProtection: Optional[bool]
    TemperatureBelowDefrostUsed: Optional[unit.DegreeC]
    NoSummerBypass: Optional[bool]

    # -- HP Water Heater
    HPWH_EF: Optional[unit._Percentage]

    # -- Water Storage
    QauntityWS: Optional[int]
    SolarThermalStorageCapacity: Optional[unit.Liter]
    StorageLossesStandby: Optional[unit.Watts_per_DegreeK]
    TotalSolarThermalStorageLosses: Optional[unit.Watts_per_DegreeK]
    InputOption: Optional[int]
    AverageHeatReleaseStorage: Optional[unit.Watts]
    TankRoomTemp: Optional[unit.DegreeC]
    TypicalStorageWaterTemperature: Optional[unit.DegreeC]

    # -- PV System
    SelectionLocation: Optional[int]
    SelectionOnSiteUtilization: Optional[int]
    SelectionUtilization: Optional[int]
    ArraySizePV: Optional[float]
    PhotovoltaicRenewableEnergy: Optional[unit.kWh]
    OnsiteUtilization: Optional[unit._Percentage]

    # -- Boilers
    EnergySourceBoilerType: Optional[int]
    CondensingBoiler: Optional[bool]
    BoilerEfficiency30: Optional[unit._Percentage]
    BoilerEfficiencyNominalOutput: Optional[float]
    AverageReturnTemperatureMeasured30Load: Optional[float]
    AverageBoilerTemperatureDesign70_55: Optional[float]
    AverageBoilerTemperatureDesign55_45: Optional[float]
    AverageBoilerTemperatureDesign35_28: Optional[float]
    StandbyHeatLossBoiler70: Optional[float]
    MaximalBoilerPower: Optional[float]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class DHW_Parameters(BaseModel):
    CoverageWithinSystem: unit._Percentage
    Unit: float
    Selection: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Heating_Parameters(BaseModel):
    CoverageWithinSystem: unit._Percentage
    Unit: float
    Selection: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Device(BaseModel):
    Name: str
    IdentNr: int
    SystemType: int
    TypeDevice: int

    UsedFor_Heating: bool
    UsedFor_DHW: bool
    UsedFor_Cooling: bool
    UsedFor_Ventilation: bool
    UsedFor_Humidification: bool
    UsedFor_Dehumidification: bool

    PH_Parameters: Optional[PH_Parameters]
    DHW_Parameters: Optional[DHW_Parameters]
    Heating_Parameters: Optional[Heating_Parameters]

    HeatRecovery: Optional[unit._Percentage]
    MoistureRecovery: Optional[unit._Percentage]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class System(BaseModel):
    Name: str
    Type: int
    IdentNr: int
    ZonesCoverage: List[ZoneCoverage]
    PHDistribution: PHDistribution
    Devices: List[Device]

    @validator("ZonesCoverage", pre=True)
    def unpack_zone_coverage(cls, v):
        return v

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Systems(BaseModel):
    Systems: List[System]


# ------------------------------------------------------------------------------
# -- Foundations


class FoundationInterface(BaseModel):
    # -- Common Attributes
    Name: str
    SettingFloorSlabType: int
    FloorSlabType: int

    # -- Heated Basement
    FloorSlabArea: Optional[unit.M2]
    U_ValueBasementSlab: Optional[unit.Watts_per_M2K]
    FloorSlabPerimeter: Optional[unit.M]
    U_ValueBasementWall: Optional[unit.Watts_per_M2K]
    DepthBasementBelowGroundSurface: Optional[unit.M]

    # -- Unheated Basement
    HeightBasementWallAboveGrade: Optional[unit.M]
    FloorCeilingArea: Optional[unit.M]
    U_ValueCeilingToUnheatedCellar: Optional[unit.Watts_per_M2K]
    U_ValueWallAboveGround: Optional[unit.Watts_per_M2K]
    BasementVolume: Optional[unit.M3]

    # -- Slab on Grade
    PositionPerimeterInsulation: Optional[int]
    PerimeterInsulationWidthDepth: Optional[unit.M]
    ConductivityPerimeterInsulation: Optional[unit.Watts_per_MK]
    ThicknessPerimeterInsulation: Optional[unit.M]

    # -- Vented Crawlspace
    U_ValueCrawlspaceFloor: Optional[unit.Watts_per_M2K]
    CrawlspaceVentOpenings: Optional[unit.M2]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Variants


class InternalGainsAdditionalData(BaseModel):
    EvaporationHeatPerPerson: unit.Watts
    HeatLossFluschingWC: bool
    QuantityWCs: unit._Int
    RoomCategory: int
    UseDefaultValuesSchool: bool
    MarginalPerformanceRatioDHW: Optional[unit._Percentage]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class PH_Building(BaseModel):
    IdentNr: int
    BuildingCategory: int
    OccupancyTypeResidential: Optional[int]
    OccupancyTypeNonResidential: Optional[int]
    BuildingStatus: int
    BuildingType: int
    OccupancySettingMethod: int
    NumberUnits: Optional[unit._Int]
    CountStories: Optional[int]
    EnvelopeAirtightnessCoefficient: unit.M3_per_Hour_per_M2
    SummerHRVHumidityRecovery: float
    FoundationInterfaces: Optional[List[FoundationInterface]]
    InternalGainsAdditionalData: Optional[InternalGainsAdditionalData]
    MechanicalRoomTemperature: unit.DegreeC
    IndoorTemperature: unit.DegreeC
    OverheatingTemperatureThreshold: unit.DegreeC
    NonCombustibleMaterials: bool

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class PassivehouseData(BaseModel):
    PH_CertificateCriteria: int
    PH_SelectionTargetData: int
    AnnualHeatingDemand: unit.kWh_per_M2
    AnnualCoolingDemand: unit.kWh_per_M2
    PeakHeatingLoad: unit.Watts_per_M2
    PeakCoolingLoad: unit.Watts_per_M2
    PH_Buildings: List[PH_Building]
    UseWUFIMeanMonthShading: bool

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Plugin(BaseModel):
    InsertPlugIn: Optional[bool]
    Name_dll: Optional[Any]
    StatusPlugIn: Optional[Any]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Variant(BaseModel):
    IdentNr: int
    Name: Optional[str]
    Remarks: Optional[str]
    PlugIn: Optional[Plugin]
    Graphics_3D: Graphics_3D
    Building: Building
    ClimateLocation: ClimateLocation
    PassivehouseData: PassivehouseData
    HVAC: Systems

    @validator("HVAC", pre=True)
    def unpack_hvac(cls, v):
        return v

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Constructions


class Material(BaseModel):
    Name: str
    ThermalConductivity: unit.Watts_per_MK
    BulkDensity: unit.KG_per_M3
    Porosity: unit._Percentage
    HeatCapacity: unit.Joule_per_KGK
    WaterVaporResistance: unit.WUFI_Vapor_Resistance_Factor
    ReferenceWaterContent: unit.KG_per_M3

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Layer(BaseModel):
    Thickness: unit.M
    Material: Material

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class Assembly(BaseModel):
    IdentNr: int
    Name: str
    Order_Layers: int
    Grid_Kind: int
    Layers: List[Layer]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class WindowType(BaseModel):
    IdentNr: int
    Name: str
    Uw_Detailed: bool
    GlazingFrameDetailed: bool
    FrameFactor: unit._Percentage
    U_Value: unit.Watts_per_M2K
    U_Value_Glazing: unit.Watts_per_M2K
    MeanEmissivity: unit._Percentage
    g_Value: unit._Percentage
    SHGC_Hemispherical: unit._Percentage
    U_Value_Frame: unit.Watts_per_M2K

    Frame_Width_Left: unit.M
    Frame_Psi_Left: unit.Watts_per_MK
    Frame_U_Left: unit.Watts_per_M2K
    Glazing_Psi_Left: unit.Watts_per_MK

    Frame_Width_Right: unit.M
    Frame_Psi_Right: unit.Watts_per_MK
    Frame_U_Right: unit.Watts_per_M2K
    Glazing_Psi_Right: unit.Watts_per_MK

    Frame_Width_Top: unit.M
    Frame_Psi_Top: unit.Watts_per_MK
    Frame_U_Top: unit.Watts_per_M2K
    Glazing_Psi_Top: unit.Watts_per_MK

    Frame_Width_Bottom: unit.M
    Frame_Psi_Bottom: unit.Watts_per_MK
    Frame_U_Bottom: unit.Watts_per_M2K
    Glazing_Psi_Bottom: unit.Watts_per_MK

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class SolarProtectionType(BaseModel):
    IdentNr: int
    Name: str
    OperationMode: int
    MaxRedFactorRadiation: Optional[unit._Percentage]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Patterns (Schedules)


class UtilizationPatternVent(BaseModel):
    Name: str
    IdentNr: int
    OperatingDays: float
    OperatingWeeks: float
    Maximum_DOS: float
    Maximum_PDF: float
    Standard_DOS: float
    Standard_PDF: float
    Basic_DOS: float
    Basic_PDF: float
    Minimum_DOS: float
    Minimum_PDF: float

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class UtilizationPattern(BaseModel):
    IdentNr: int
    Name: str
    HeightUtilizationLevel: float
    BeginUtilization: unit.Hour
    EndUtilization: unit.Hour
    AnnualUtilizationDays: unit.Days_per_Year
    IlluminationLevel: unit.Lux
    RelativeAbsenteeism: unit._Percentage
    PartUseFactorPeriodForLighting: unit._Percentage

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Project


class Date_Project(BaseModel):
    Year: int
    Month: int
    Day: int
    Hour: int
    Minutes: int

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


class ProjectData(BaseModel):
    Year_Construction: int
    OwnerIsClient: bool
    WhiteBackgroundPictureBuilding: Optional[bool]
    Customer_Name: Optional[str]
    Customer_Street: Optional[str]
    Customer_Locality: Optional[str]
    Customer_PostalCode: Optional[str]
    Customer_Tel: Optional[str]
    Customer_Email: Optional[str]
    Building_Name: Optional[str]
    Building_Street: Optional[str]
    Building_Locality: Optional[str]
    Building_PostalCode: Optional[str]
    Owner_Name: Optional[str]
    Owner_Street: Optional[str]
    Owner_Locality: Optional[str]
    Owner_PostalCode: Optional[str]
    Responsible_Name: Optional[str]
    Responsible_Street: Optional[str]
    Responsible_Locality: Optional[str]
    Responsible_PostalCode: Optional[str]
    Responsible_Tel: Optional[str]
    Responsible_LicenseNr: Optional[str]
    Responsible_Email: Optional[str]
    Date_Project: Optional[Date_Project]

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)


# ------------------------------------------------------------------------------
# -- Root


class WUFIplusProject(BaseModel):
    DataVersion: int
    UnitSystem: int
    ProgramVersion: str
    Scope: int
    DimensionsVisualizedGeometry: int
    ProjectData: ProjectData
    UtilizationPatternsPH: List[UtilizationPattern]
    UtilisationPatternsVentilation: List[UtilizationPatternVent]
    WindowTypes: List[WindowType]
    Assemblies: List[Assembly]
    Variants: List[Variant]
    SolarProtectionTypes: Optional[List[SolarProtectionType]]

    @validator("UnitSystem", pre=True)
    def unpack_unit_system(cls, v):
        """
        Since the model gets converted to SI units, always set this to 1, no matter what.
        """
        return 1

    _unpack_xml_tag_name = validator("*", allow_reuse=True, pre=True)(unpack_xml_tag)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "DHW+Distribution" worksheet."""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any

from ph_units.unit_type import Unit

from PHX.xl import xl_app
from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import hot_water_tank, hot_water_piping


class RecircPiping:
    """The Recirculation Piping Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self._header_row: Optional[int] = None

    @property
    def header_row(self) -> int:
        if not self._header_row:
            self._header_row = self.find_header_row()
        return self._header_row

    def find_header_row(self, _row_start: int = 100, _rows: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.recirc_piping.locator_col_entry,
            _row_start=_row_start,
            _row_end=_row_start + _rows,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if self.shape.recirc_piping.locator_string_entry == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.recirc_piping.locator_string_entry}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.recirc_piping.locator_string_entry}?"
        )


class BranchPiping:
    """The Branch Piping Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self._header_row: Optional[int] = None

    @property
    def header_row(self) -> int:
        if not self._header_row:
            self._header_row = self.find_header_row()
        return self._header_row

    def find_header_row(self, _row_start: int = 100, _rows: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.branch_piping.locator_col_entry,
            _row_start=_row_start,
            _row_end=_row_start + _rows,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if self.shape.branch_piping.locator_string_entry == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.branch_piping.locator_string_entry}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.branch_piping.locator_string_entry}?"
        )


class DHWPiping:
    """The DHW Piping Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self._header_row: Optional[int] = None
        self.recirc_piping = RecircPiping(self.xl, self.shape)
        self.branch_piping = BranchPiping(self.xl, self.shape)

    def find_header_row(self, _row_start: int = 100, _row_end: int = 200) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.branch_piping.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if self.shape.branch_piping.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.branch_piping.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.branch_piping.locator_string_header}?"
        )

    @property
    def header_row(self) -> int:
        if not self._header_row:
            self._header_row = self.find_header_row()
        return self._header_row

    def find_recirc_piping_start_row(self) -> int:
        return self.recirc_piping.header_row

    def find_branch_piping_start_row(self) -> int:
        return self.branch_piping.header_row


@dataclass
class TankData:
    """Convenience Wrapper for Tank Data read in from the PHPP."""

    type: str = ""
    heat_loss_rate: Unit = field(default_factory=Unit)
    volume: Unit = field(default_factory=Unit)

    @classmethod
    def from_phpp_data(cls, _d: Dict[str, Any]) -> "TankData":
        return cls(
            type=_d["type"],
            heat_loss_rate=Unit(_d["heat_loss_rate"], _d["heat_loss_rate_unit"]),
            volume=Unit(_d["volume"], _d["volume_unit"]),
        )


class Tank:
    """An individual tank entry item."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self._entry_row_start: Optional[int] = None

    @property
    def entry_row_start(self) -> int:
        """The row where the tank entry starts."""
        if not self._entry_row_start:
            self._entry_row_start = self.find_entry_row_start()
        return self._entry_row_start

    def find_entry_row_start(self) -> int:
        """Find the row where the tank entry starts."""
        return self.shape.tanks.entry_row_start

    def get_phpp_data(self, _tank_num: int) -> TankData:
        """Get the PHPP data for the specified tank number."""
        shape = self.shape.tanks
        rows = shape.input_rows
        col = getattr(shape.input_columns, f"tank_{_tank_num}")

        phpp_data = {
            "type": self.xl.get_data(
                self.shape.name,
                f"{col}{self.entry_row_start + rows.tank_type.row}",
            ),
            "heat_loss_rate": self.xl.get_data(
                self.shape.name,
                f"{col}{self.entry_row_start + rows.standby_losses.row}",
            ),
            "heat_loss_rate_unit": rows.standby_losses.unit,
            "volume": self.xl.get_data(
                self.shape.name,
                f"{col}{self.entry_row_start + rows.storage_capacity.row}",
            ),
            "volume_unit": rows.storage_capacity.unit,
        }

        return TankData.from_phpp_data(phpp_data)


class Tanks:
    """The Tanks (Storage Heat Loss) Section Group"""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw) -> None:
        self.xl = _xl
        self.shape = _shape
        self._header_row: Optional[int] = None

        self.tank_1 = Tank(self.xl, self.shape)
        self.tank_2 = Tank(self.xl, self.shape)

    @property
    def header_row(self) -> int:
        """The row where the tank header starts."""
        if not self._header_row:
            self._header_row = self.find_header_row()
        return self._header_row

    def find_header_row(self, _row_start: int = 150, _row_end: int = 200) -> int:
        """Find the row where the tank header starts."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.tanks.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if self.shape.tanks.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.tanks.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.tanks.locator_string_header}?"
        )

    def get_all_tank_device_data(self) -> List[TankData]:
        """Get all the tank data from the spreadsheet."""
        return [
            self.tank_1.get_phpp_data(1),
            self.tank_2.get_phpp_data(2),
        ]


class HotWater:
    """IO Controller for the PHPP 'DHW+Distribution' PHPP worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Dhw):
        self.xl = _xl
        self.shape = _shape
        self.tanks = Tanks(self.xl, self.shape)
        self.dhw_piping = DHWPiping(self.xl, self.shape)

    def write_tanks(self, _phpp_hw_tanks: List[hot_water_tank.TankInput]) -> None:
        """Write the tank data to the spreadsheet."""
        for phpp_Tank_input in _phpp_hw_tanks:
            for item in phpp_Tank_input.create_xl_items(
                self.shape.name, self.tanks.tank_1.entry_row_start
            ):
                self.xl.write_xl_item(item)

    def write_branch_piping(
        self, _phpp_branch_piping: List[hot_water_piping.BranchPipingInput]
    ) -> None:
        """Write the branch piping data to the spreadsheet."""
        for pipe_inputs in _phpp_branch_piping:
            for item in pipe_inputs.create_xl_items(
                self.shape.name, self.dhw_piping.branch_piping.header_row
            ):
                self.xl.write_xl_item(item)

    def write_recirc_piping(
        self, _phpp_recirc_piping: List[hot_water_piping.BranchPipingInput]
    ) -> None:
        """Write the recirc piping data to the spreadsheet."""
        for pipe_inputs in _phpp_recirc_piping:
            for item in pipe_inputs.create_xl_items(
                self.shape.name, self.dhw_piping.recirc_piping.header_row
            ):
                self.xl.write_xl_item(item)

    def get_all_tank_device_data(self) -> List[TankData]:
        """Get all the tank data from the PHPP worksheet."""
        return self.tanks.get_all_tank_device_data()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for building PHX-Geometry from Ladybug / Honeybee Geometry."""

from typing import Union

from honeybee import aperture, face, shade
from ladybug_geometry.geometry3d.pointvector import Vector3D, Point3D
from ladybug_geometry.geometry3d.plane import Plane
from PHX.model import geometry


class InvalidRectangularFaceError(Exception):
    def __init__(self, _hb_face: Union[aperture.Aperture, face.Face, shade.Shade]):
        self.msg = (
            f"Error: Cannot create a PhxPolygonRectangular for the"
            f"Honeybee Face: {_hb_face.display_name}. Must have at least 4 vertices?"
            f"Only {len(_hb_face.vertices)} vertices found?"
        )
        super().__init__(self.msg)


# -----------------------------------------------------------------------------


def create_PhxVertix_from_lbt_Point3D(_lbt_Point3D: Point3D) -> geometry.PhxVertix:
    """Returns a new PhxVertix object with attributes based on an LBT-Point3D."""
    return geometry.PhxVertix(_lbt_Point3D.x, _lbt_Point3D.y, _lbt_Point3D.z)


def create_PhxVector_from_lbt_Vector3D(_lbt_vector3d: Vector3D) -> geometry.PhxVector:
    """Return a new PhxVector with attributes based on an LBT-Vector3D."""
    return geometry.PhxVector(
        _lbt_vector3d.x,
        _lbt_vector3d.y,
        _lbt_vector3d.z,
    )


def create_PhxPlane_from_lbt_Plane(_lbt_plane: Plane) -> geometry.PhxPlane:
    """Return a new PhxPlane with attributes based on an LBT-Plane"""
    return geometry.PhxPlane(
        create_PhxVector_from_lbt_Vector3D(_lbt_plane.n),
        create_PhxVertix_from_lbt_Point3D(_lbt_plane.o),
        create_PhxVector_from_lbt_Vector3D(_lbt_plane.x),
        create_PhxVector_from_lbt_Vector3D(_lbt_plane.y),
    )


def create_PhxPolygon_from_hb_Face(
    _hb_face: Union[aperture.Aperture, face.Face, shade.Shade]
) -> geometry.PhxPolygon:
    """Return a new PhxPolygon based on an input honeybee-Face.

    Arguments:
    ----------
        * _hb_face (Union[aperture.Aperture, face.Face, shade.Shade]): The Honeybee-Face
            to base the PhxPolygon on.

    Returns:
    --------
        * (geometry.PhxPolygon): The new PhxPolygon object.
    """
    phx_polygon = geometry.PhxPolygon(
        _hb_face.display_name,
        _hb_face.geometry.area,
        create_PhxVertix_from_lbt_Point3D(_hb_face.geometry.center),
        create_PhxVector_from_lbt_Vector3D(_hb_face.normal),
        create_PhxPlane_from_lbt_Plane(_hb_face.geometry.plane),
    )

    for v in _hb_face.vertices:
        phx_polygon.add_vertix(create_PhxVertix_from_lbt_Point3D(v))

    return phx_polygon


def create_PhxPolygonRectangular_from_hb_Face(
    _hb_face: Union[aperture.Aperture, face.Face, shade.Shade], _tolerance: float = 0.001
) -> geometry.PhxPolygonRectangular:
    """Return a new PhxPolygonRectangular based on an input honeybee-Face.

    Arguments:
    ----------
        * _hb_face (Union[aperture.Aperture, face.Face, shade.Shade]): The Honeybee-Face
            to base the PhxPolygon on.

    Returns:
    --------
        * (geometry.PhxPolygonRectangular): The new PhxPolygonRectangular object.
    """

    # -- Try and clean up the polygon geometry
    face_polygon = _hb_face.geometry
    face_polygon = face_polygon.remove_colinear_vertices(_tolerance)
    if len(face_polygon.vertices) != 4:
        raise InvalidRectangularFaceError(_hb_face)

    # -- Create a new PhxPolygonRectangular
    phx_polygon = geometry.PhxPolygonRectangular(
        _hb_face.display_name,
        _hb_face.geometry.area,
        _center=create_PhxVertix_from_lbt_Point3D(_hb_face.geometry.center),
        normal_vector=create_PhxVector_from_lbt_Vector3D(_hb_face.normal),
        plane=create_PhxPlane_from_lbt_Plane(_hb_face.geometry.plane),
    )

    # -- Set the vertices using the upper_left_counter_clockwise_vertices method.
    # -- Note: cannot use the 'lower_right_corner' property as it is not always
    # -- the lower right corner of the face. It shows the vertix of the bounding box
    # -- which is not aligned when the face has been rotated.
    vertices = face_polygon.upper_left_counter_clockwise_vertices
    phx_polygon.vertix_upper_left = create_PhxVertix_from_lbt_Point3D(vertices[0])
    phx_polygon.vertix_lower_left = create_PhxVertix_from_lbt_Point3D(vertices[1])
    phx_polygon.vertix_lower_right = create_PhxVertix_from_lbt_Point3D(vertices[2])
    phx_polygon.vertix_upper_right = create_PhxVertix_from_lbt_Point3D(vertices[3])

    return phx_polygon



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Construction, Materials Classes"""

from __future__ import annotations
from typing import ClassVar, Union, List, Optional, Generator
from dataclasses import dataclass, field
import uuid


@dataclass
class PhxMaterial:
    display_name: str = ""
    conductivity: float = 0.0
    density: float = 0.0
    porosity: float = 0.0
    heat_capacity: float = 0.0
    water_vapor_resistance: float = 0.0
    reference_water: float = 0.0
    percentage_of_assembly: float = 1.0


@dataclass
class PhxLayer:
    thickness_m: float = 0.0
    materials: List[PhxMaterial] = field(default_factory=list)

    @property
    def material(self) -> PhxMaterial:
        """Return the first PhxMaterial from the self.materials collection."""
        if not self.materials:
            # -- Return a default PhxMaterial if no materials are set.
            return PhxMaterial()
        else:
            # -- Otherwise return the first material in the collection.
            return self.materials[0]

    def add_material(self, _material: PhxMaterial) -> None:
        self.materials.append(_material)

    @property
    def thickness_mm(self):
        """Returns the thickness of the layer in MM"""
        return self.thickness_m * 1000

    @classmethod
    def from_total_u_value(cls, _total_u_value: float) -> PhxLayer:
        """Returns a new PhxLayer with a single material with the given U-Value.
        note that this will assign a default thickness of 1m to the layer.
        """
        obj = cls()
        obj.thickness_m = 1

        new_mat = PhxMaterial()
        new_mat.conductivity = obj.thickness_m * _total_u_value
        new_mat.display_name = "Material"
        obj.materials.append(new_mat)

        return obj


@dataclass
class PhxConstructionOpaque:
    _count: ClassVar[int] = 0

    _identifier: Union[uuid.UUID, str] = field(init=False, default_factory=uuid.uuid4)
    id_num: int = field(init=False, default=0)
    display_name: str = ""
    layer_order: int = 2  # Outside to Inside
    grid_kind: int = 2  # Medium
    layers: list[PhxLayer] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def identifier(self) -> str:
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: str):
        if not _in:
            return
        self._identifier = str(_in)

    def __hash__(self) -> int:
        return hash(self.identifier)

    @classmethod
    def from_total_u_value(
        cls, _total_u_value: float, _display_name: str = ""
    ) -> PhxConstructionOpaque:
        """Returns a new PhxConstructionOpaque with a single layer with the given U-Value."""
        obj = cls()
        obj.display_name = _display_name or f"U-Value: {_total_u_value}"
        obj.layers.append(PhxLayer.from_total_u_value(_total_u_value))
        return obj


# ------------------------------------------------------------
# Windows


@dataclass
class PhxWindowFrameElement:
    width: float = 0.1  # m
    u_value: float = 1.0  # W/m2k
    psi_glazing: float = 0.00  # W/mk
    psi_install: float = 0.00  # W/mk

    @classmethod
    def from_total_u_value(cls, _total_u_value: float) -> PhxWindowFrameElement:
        """Returns a new PhxWindowFrameElement with u-values set to a single value."""
        obj = cls()
        obj.width = 0.1  # m
        obj.u_value = _total_u_value
        obj.psi_glazing = 0.0
        obj.psi_install = 0.0
        return obj


@dataclass
class PhxConstructionWindow:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    _identifier: Union[uuid.UUID, str] = field(init=False, default_factory=uuid.uuid4)
    display_name: str = ""
    _glazing_type_display_name: str = ""
    _frame_type_display_name: str = ""

    use_detailed_uw: bool = True
    use_detailed_frame: bool = True

    u_value_window: float = 1.0
    u_value_glass: float = 1.0
    u_value_frame: float = 1.0

    frame_top: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_right: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_bottom: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_left: PhxWindowFrameElement = field(default_factory=PhxWindowFrameElement)
    frame_factor: float = 0.75

    glass_mean_emissivity: float = 0.1
    glass_g_value: float = 0.4

    _id_num_shade: int = -1

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def glazing_type_display_name(self) -> str:
        return self._glazing_type_display_name or self.display_name

    @glazing_type_display_name.setter
    def glazing_type_display_name(self, _in: str) -> None:
        self._glazing_type_display_name = _in

    @property
    def frame_type_display_name(self) -> str:
        return self._frame_type_display_name or self.display_name

    @frame_type_display_name.setter
    def frame_type_display_name(self, _in: str) -> None:
        self._frame_type_display_name = _in

    @property
    def identifier(self) -> str:
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: str) -> None:
        if not _in:
            return
        self._identifier = str(_in)

    @property
    def id_num_shade(self) -> int:
        return self._id_num_shade

    @id_num_shade.setter
    def id_num_shade(self, _in: Optional[int]) -> None:
        if _in != None:
            self._id_num_shade = _in

    @classmethod
    def from_total_u_value(
        cls, _total_u_value: float, _g_value: float = 0.4, _display_name: str = ""
    ) -> PhxConstructionWindow:
        """Returns a new PhxConstructionWindow with all u-values set to a single value."""
        obj = cls()
        obj.display_name = _display_name or f"U-Value: {_total_u_value}"
        obj.frame_type_display_name = _display_name or f"U-Value: {_total_u_value}"
        obj.glazing_type_display_name = _display_name or f"U-Value: {_total_u_value}"

        obj.u_value_window = _total_u_value
        obj.u_value_glass = _total_u_value
        obj.u_value_frame = _total_u_value

        obj.glass_g_value = _g_value

        obj.frame_top = PhxWindowFrameElement.from_total_u_value(_total_u_value)
        obj.frame_right = PhxWindowFrameElement.from_total_u_value(_total_u_value)
        obj.frame_bottom = PhxWindowFrameElement.from_total_u_value(_total_u_value)
        obj.frame_left = PhxWindowFrameElement.from_total_u_value(_total_u_value)

        return obj

    @property
    def frames(self) -> Generator[PhxWindowFrameElement, None, None]:
        """Returns a generator of all frame elements in order: Top, Right, Bottom, Left."""
        for frame in (
            self.frame_top,
            self.frame_right,
            self.frame_bottom,
            self.frame_left,
        ):
            yield frame

    def set_all_frames_u_value(self, _u_value: float) -> None:
        """Sets the u-value of all frame elements to the given value."""
        for frame in self.frames:
            frame.u_value = _u_value

    def set_all_frames_width(self, _width: float) -> None:
        """Sets the width of all frame elements to the given value."""
        for frame in self.frames:
            frame.width = _width

    def set_all_frames_psi_glazing(self, _psi_glazing: float) -> None:
        """Sets the psi_glazing of all frame elements to the given value."""
        for frame in self.frames:
            frame.psi_glazing = _psi_glazing

    def set_all_frames_psi_install(self, _psi_install: float) -> None:
        """Sets the psi_install of all frame elements to the given value."""
        for frame in self.frames:
            frame.psi_install = _psi_install



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Areas / Surface-Entry row"""

from dataclasses import dataclass
from functools import partial
import re
from typing import List, Optional, Dict

from PHX.model import components, geometry
from PHX.model.enums.building import ComponentExposureExterior, ComponentFaceType

from PHX.xl import xl_data
from PHX.PHPP.phpp_localization import shape_model


@dataclass
class SurfaceRow:
    """A single Areas/Surface entry row."""

    __slots__ = ("shape", "phx_polygon", "phx_component", "phpp_assembly_id_name")
    shape: shape_model.Areas
    phx_polygon: geometry.PhxPolygon
    phx_component: components.PhxComponentOpaque
    phpp_assembly_id_name: Optional[str]

    @property
    def phpp_group_number(self) -> int:
        """Return the correct PHPP 'Group Number' depending on the exposure and type."""

        if self.phx_component.exposure_exterior == ComponentExposureExterior.SURFACE:
            return 18
        elif self.phx_component.face_type == ComponentFaceType.WALL:
            if self.phx_component.exposure_exterior == ComponentExposureExterior.EXTERIOR:
                return 8
            else:
                return 9
        elif self.phx_component.face_type == ComponentFaceType.FLOOR:
            return 11
        elif self.phx_component.face_type == ComponentFaceType.ROOF_CEILING:
            return 10
        else:
            return 12

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f"{getattr(self.shape.surface_rows.inputs, _field_name).column}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.surface_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAreas = partial(xl_data.XlItem, _sheet_name)
        return [
            XLItemAreas(create_range("description"), f"'{self.phx_polygon.display_name}"),
            XLItemAreas(create_range("group_number"), self.phpp_group_number),
            XLItemAreas(create_range("quantity"), 1),
            XLItemAreas(
                create_range("area"),
                self.phx_polygon.area,
                "M2",
                self._get_target_unit("area"),
            ),
            XLItemAreas(create_range("assembly_id"), self.phpp_assembly_id_name),
            XLItemAreas(
                create_range("orientation"), self.phx_polygon.cardinal_orientation_angle
            ),
            XLItemAreas(create_range("angle"), self.phx_polygon.angle_from_horizontal),
            XLItemAreas(create_range("shading"), 0.5),
            XLItemAreas(create_range("absorptivity"), 0.6),
            XLItemAreas(create_range("emissivity"), 0.9),
        ]


def get_name_from_assembly_id(
    _phpp_assembly_id: Optional[xl_data.xl_range_single_value],
) -> str:
    """Return the face's construction PHPP-Name (ie: "MyConst") from id-string."""
    try:
        return str(_phpp_assembly_id).split("-", 1)[1]
    except:
        if _phpp_assembly_id == None or _phpp_assembly_id == "None":
            return ""
        else:
            msg = (
                f"Error getting construction PHPP-Name? "
                f"Could not split {_phpp_assembly_id} on '-'?"
            )
            raise Exception(msg)


@dataclass
class ExistingSurfaceRow:
    """The data from an existing PHPP Surface Entry row."""

    __slots__ = ("shape", "data", "group_type_exposure_map")
    shape: shape_model.Areas
    data: list
    group_type_exposure_map: Dict[int, str]  # From io_areas.Areas() parent class

    @property
    def is_empty(self) -> bool:
        if self.name in ["-", "", "None", None]:
            return True
        if self.face_group_type_phpp_string in ["-", "", "None", None]:
            return True
        if self.face_group_type_phpp_number == 0:
            return True
        if self.face_type == ComponentFaceType.NONE:
            return True
        if self.face_exposure == ComponentExposureExterior.NONE:
            return True
        return False

    @property
    def name(self) -> str:
        col_letter = self.shape.surface_rows.inputs.description.column
        col_number_as_index = xl_data.xl_ord(str(col_letter)) - 65
        return str(self.data[col_number_as_index])

    @property
    def face_group_type_phpp_string(self) -> str:
        """Return the face's Group-Type string from the row-data."""
        col_letter = self.shape.surface_rows.inputs.group_number.column
        col_number_as_index = xl_data.xl_ord(str(col_letter)) - 65
        return str(self.data[col_number_as_index])

    @property
    def face_group_type_phpp_number(self) -> int:
        """Return the face's Group-Type number as an int from the Group-Type string."""
        phpp_data = self.face_group_type_phpp_string
        if phpp_data in ["-", "", "None", None]:
            return 0

        result = re.split(r"\D+", phpp_data, 2)
        if not result:
            msg = (
                f"Error getting Group-Type number? Could not find a number "
                f"in the Group-Type string from PHPP: '{phpp_data}'?"
            )
            raise Exception(msg)

        try:
            return int(result[0])
        except:
            msg = (
                f"Error getting Group-Type number? Could not convert '{result[0]}' "
                f"to int from the PHPP value: '{phpp_data}' or type: {type(phpp_data)}?"
            )
            raise Exception(msg)

    @property
    def face_exposure_phpp_letter(self) -> str:
        """Return the face's exposure-type letter ("A", "B", etc..) based on the group_type_phpp_number"""
        return self.group_type_exposure_map[self.face_group_type_phpp_number]

    @property
    def face_type(self) -> ComponentFaceType:
        """Return the face type enum (WALL, FLOOR, etc..) based on the group_type_phpp_number."""
        type_map = {
            0: ComponentFaceType.NONE,
            2: ComponentFaceType.WINDOW,
            3: ComponentFaceType.WINDOW,
            4: ComponentFaceType.WINDOW,
            5: ComponentFaceType.WINDOW,
            6: ComponentFaceType.WINDOW,
            7: ComponentFaceType.WINDOW,
            8: ComponentFaceType.WALL,
            9: ComponentFaceType.WALL,
            10: ComponentFaceType.ROOF_CEILING,
            11: ComponentFaceType.FLOOR,
            12: ComponentFaceType.CUSTOM,
            13: ComponentFaceType.CUSTOM,
            14: ComponentFaceType.CUSTOM,
            18: ComponentFaceType.ADIABATIC,
        }
        return type_map[self.face_group_type_phpp_number]

    @property
    def face_exposure(self) -> ComponentExposureExterior:
        """Return the exposure type enum (EXTERIOR, GROUND, SURFACE) based on the face_exposure_phpp_letter"""
        type_map = {
            "A": ComponentExposureExterior.EXTERIOR,
            "B": ComponentExposureExterior.GROUND,
            "": ComponentExposureExterior.NONE,
        }
        return type_map.get(
            self.face_exposure_phpp_letter, ComponentExposureExterior.SURFACE
        )

    @property
    def face_construction_phpp_id(self) -> str:
        """Return the face's construction PHPP-ID (ie: "01ud-MyConst") from the row-data."""
        col_letter = self.shape.surface_rows.inputs.assembly_id.column
        col_number_as_index = xl_data.xl_ord(str(col_letter)) - 65
        return str(self.data[col_number_as_index])

    @property
    def face_construction_phpp_name(self) -> str:
        """Return the face's construction PHPP-Name (ie: "MyConst") from the row-data."""
        return get_name_from_assembly_id(self.face_construction_phpp_id)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Passive House Certification Classes"""

from __future__ import annotations
from typing import ClassVar, Optional
from dataclasses import dataclass, field
from enum import Enum

from PHX.model import ground
from PHX.model.enums import phi_certification_phpp_9, phius_certification
from PHX.model.enums.hvac import PhxSummerBypassMode


@dataclass
class PhxSetpoints:
    winter: float = 20.0  # deg. C
    summer: float = 25.0  # deg. C

    def __eq__(self, other: PhxSetpoints) -> bool:
        TOLERANCE = 0.001
        if abs(self.winter - other.winter) > TOLERANCE:
            return False
        if abs(self.summer - other.summer) > TOLERANCE:
            return False
        return True


@dataclass
class PhxPhBuildingData:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)

    num_of_units: Optional[int] = 1
    num_of_floors: Optional[int] = 1
    occupancy_setting_method: int = 2  # Design
    airtightness_q50: float = 1.0  # m3/hr-m2-envelope
    airtightness_n50: float = 1.0  # ach
    wind_coefficient_e: float = 0.07
    wind_coefficient_f: float = 15
    setpoints: PhxSetpoints = field(default_factory=PhxSetpoints)
    mech_room_temp: float = 20.0
    non_combustible_materials: bool = False
    foundations: list[ground.PhxFoundation] = field(default_factory=list)
    summer_hrv_bypass_mode: PhxSummerBypassMode = PhxSummerBypassMode.ALWAYS

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def add_foundation(self, _input: ground.PhxFoundation) -> None:
        self.foundations.append(_input)

    def __eq__(self, other: PhxPhBuildingData) -> bool:
        TOLERANCE = 0.001
        if self.num_of_units != other.num_of_units:
            return False
        if self.num_of_floors != other.num_of_floors:
            return False
        if self.occupancy_setting_method != other.occupancy_setting_method:
            return False
        if abs(self.airtightness_q50 - other.airtightness_q50) > TOLERANCE:
            return False
        if abs(self.airtightness_n50 - other.airtightness_n50) > TOLERANCE:
            return False
        if abs(self.wind_coefficient_e - other.wind_coefficient_e) > TOLERANCE:
            return False
        if abs(self.wind_coefficient_f - other.wind_coefficient_f) > TOLERANCE:
            return False
        if self.setpoints != other.setpoints:
            return False
        if abs(self.mech_room_temp - other.mech_room_temp) > TOLERANCE:
            return False
        if self.non_combustible_materials != other.non_combustible_materials:
            return False
        if self.foundations != other.foundations:
            return False
        if self.summer_hrv_bypass_mode != other.summer_hrv_bypass_mode:
            return False
        return True


# -----------------------------------------------------------------------------
@dataclass
class PhxPhiusCertificationCriteria:
    ph_selection_target_data: int = 2

    phius_annual_heating_demand: float = 15.0
    phius_annual_cooling_demand: float = 15.0
    phius_peak_heating_load: float = 10.0
    phius_peak_cooling_load: float = 10.0

    def __eq__(self, other: PhxPhiusCertificationCriteria) -> bool:
        TOLERANCE = 0.001
        if (
            abs(self.phius_annual_heating_demand - other.phius_annual_heating_demand)
            > TOLERANCE
        ):
            return False
        if (
            abs(self.phius_annual_cooling_demand - other.phius_annual_cooling_demand)
            > TOLERANCE
        ):
            return False
        if abs(self.phius_peak_heating_load - other.phius_peak_heating_load) > TOLERANCE:
            return False
        if abs(self.phius_peak_cooling_load - other.phius_peak_cooling_load) > TOLERANCE:
            return False
        return True


@dataclass
class PhxPhiusCertificationSettings:
    phius_building_certification_program = (
        phius_certification.PhiusCertificationProgram.PHIUS_2021_CORE
    )
    phius_building_category_type = (
        phius_certification.PhiusCertificationBuildingCategoryType.RESIDENTIAL_BUILDING
    )
    phius_building_use_type = (
        phius_certification.PhiusCertificationBuildingUseType.RESIDENTIAL
    )
    phius_building_status = (
        phius_certification.PhiusCertificationBuildingStatus.IN_PLANNING
    )
    phius_building_type = (
        phius_certification.PhiusCertificationBuildingType.NEW_CONSTRUCTION
    )

    def __eq__(self, other: PhxPhiusCertificationSettings) -> bool:
        if (
            self.phius_building_certification_program
            != other.phius_building_certification_program
        ):
            return False
        if self.phius_building_category_type != other.phius_building_category_type:
            return False
        if self.phius_building_use_type != other.phius_building_use_type:
            return False
        if self.phius_building_status != other.phius_building_status:
            return False
        if self.phius_building_type != other.phius_building_type:
            return False
        return True


@dataclass
class PhxPhiusCertification:
    phius_certification_criteria: PhxPhiusCertificationCriteria = field(
        default_factory=PhxPhiusCertificationCriteria
    )
    phius_certification_settings: PhxPhiusCertificationSettings = field(
        default_factory=PhxPhiusCertificationSettings
    )

    # TODO: Refactor this out to someplace more general than inside Phius....
    ph_building_data: PhxPhBuildingData = field(default_factory=PhxPhBuildingData)
    use_monthly_shading: bool = True

    def __eq__(self, other: PhxPhiusCertification) -> bool:
        if self.phius_certification_criteria != other.phius_certification_criteria:
            return False
        if self.phius_certification_settings != other.phius_certification_settings:
            return False
        if self.ph_building_data != other.ph_building_data:
            return False
        if self.use_monthly_shading != other.use_monthly_shading:
            return False
        return True


# -----------------------------------------------------------------------------
@dataclass
class PhxPhiCertificationSettings:
    phi_building_category_type: Enum = (
        phi_certification_phpp_9.PhiCertBuildingCategoryType.RESIDENTIAL_BUILDING
    )
    phi_building_use_type: Enum = phi_certification_phpp_9.PhiCertBuildingUseType.DWELLING
    phi_building_ihg_type: Enum = phi_certification_phpp_9.PhiCertIHGType.STANDARD
    phi_building_occupancy_type: Enum = (
        phi_certification_phpp_9.PhiCertOccupancyType.STANDARD
    )

    phi_certification_type: Enum = phi_certification_phpp_9.PhiCertType.PASSIVE_HOUSE
    phi_certification_class: Enum = phi_certification_phpp_9.PhiCertClass.CLASSIC
    phi_pe_type: Enum = phi_certification_phpp_9.PhiCertificationPEType.PER
    phi_enerphit_type: Enum = phi_certification_phpp_9.PhiCertEnerPHitType.BY_DEMAND
    phi_retrofit_type: Enum = phi_certification_phpp_9.PhiCertRetrofitType.NEW_BUILDING

    def __eq__(self, other: PhxPhiCertificationSettings) -> bool:
        if self.phi_building_category_type != other.phi_building_category_type:
            return False
        if self.phi_building_use_type != other.phi_building_use_type:
            return False
        if self.phi_building_ihg_type != other.phi_building_ihg_type:
            return False
        if self.phi_building_occupancy_type != other.phi_building_occupancy_type:
            return False
        if self.phi_certification_type != other.phi_certification_type:
            return False
        if self.phi_certification_class != other.phi_certification_class:
            return False
        if self.phi_pe_type != other.phi_pe_type:
            return False
        if self.phi_enerphit_type != other.phi_enerphit_type:
            return False
        if self.phi_retrofit_type != other.phi_retrofit_type:
            return False
        return True


@dataclass
class PhxPhiCertification:
    phi_certification_settings: PhxPhiCertificationSettings = field(
        default_factory=PhxPhiCertificationSettings
    )
    version: int = 9

    def __eq__(self, other: PhxPhiCertification) -> bool:
        if self.phi_certification_settings != other.phi_certification_settings:
            return False
        if self.version != other.version:
            return False
        return True
